###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

__author__ = "Chris Burr <c.b@cern.ch>"

import os
import re
import tempfile
import shutil
from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError, check_output, run

import pytest
from LbPlatformUtils.inspect import os_id

if os_id().endswith("slc6"):
    CAN_RUN_NATIVE_SLC5 = True
    CAN_RUN_NATIVE_SLC6 = True
    CAN_RUN_NATIVE_CENTOS7 = False
elif os_id().endswith("centos7"):
    CAN_RUN_NATIVE_SLC5 = False
    CAN_RUN_NATIVE_SLC6 = True
    CAN_RUN_NATIVE_CENTOS7 = True
elif os_id().endswith("centos8"):
    CAN_RUN_NATIVE_SLC5 = False
    CAN_RUN_NATIVE_SLC6 = False
    CAN_RUN_NATIVE_CENTOS7 = False
else:
    CAN_RUN_NATIVE_SLC5 = False
    CAN_RUN_NATIVE_SLC6 = False
    CAN_RUN_NATIVE_CENTOS7 = False
CONTAINER_IMPLEMENTATION = os.environ.get("CONTAINER_IMPLEMENTATION", "")


BASE_LBRUN_COMMAND = ["lb-run", "--debug"]
SETUPPTOJECT_COMMAND = ["--platform", "x86_64-slc5-gcc46-opt", "Gaudi/v23r0"]


def _check_gaudirun(options):
    output = check_output(BASE_LBRUN_COMMAND + options + ["gaudirun.py"], stderr=STDOUT)
    output = output.decode()
    assert "Application Manager Terminated successfully" in output
    used_container = (
        "preparing singularity wrapper command" in output
        or "preparing apptainer wrapper command" in output
    )
    return output, used_container

@pytest.fixture(scope="session", autouse=True)
def setup():
    if not os.path.isdir("/cvmfs/lhcb.cern.ch"):
        pytest.skip("Skipping integration tests due to missing siteroot")
    if not os.path.isdir("/cvmfs/cernvm-prod.cern.ch"):
        pytest.skip("Skipping integration tests due to missing cernvm")
    if not CONTAINER_IMPLEMENTATION:
        pytest.skip(
            "Skipping integration tests due CONTAINER_IMPLEMENTATION not being set"
        )
    if not shutil.which(CONTAINER_IMPLEMENTATION):
        raise RuntimeError("Failed to find %s" % CONTAINER_IMPLEMENTATION)

    if CONTAINER_IMPLEMENTATION == "singularity":
        # Only singularity 3 will work correctly
        singularity_version = check_output(
            [CONTAINER_IMPLEMENTATION, "--version"]
        ).decode()
        match = re.search(r"(?:^| )(\d+)\.\d+\.\d+", singularity_version)
        assert match, singularity_version
        if int(match.groups()[0]) < 3:
            pytest.skip("Singularity 3 or newer is required for integration tests")

    from os.path import abspath, dirname, join, pardir

    # FIXME compatibility py2-py3
    os.environ["CMAKE_PREFIX_PATH"] = abspath(
        join(
            dirname(__file__),
            pardir,
            pardir,
            pardir,
            pardir,
            pardir,
            "python",
            "LbEnv",
            "ProjectEnv",
            "tests",
            "data",
        )
    )
    yield


def test_setupproject():
    command = SETUPPTOJECT_COMMAND
    try:
        output, used_container = _check_gaudirun(["--disallow-containers"] + command)
    except CalledProcessError:
        assert not CAN_RUN_NATIVE_SLC5
    else:
        assert CAN_RUN_NATIVE_SLC5, output

    output, used_container = _check_gaudirun(
        ["--container", CONTAINER_IMPLEMENTATION] + command
    )
    assert used_container, output

    output, used_container = _check_gaudirun(command)
    assert used_container != CAN_RUN_NATIVE_SLC5, output
    if not CAN_RUN_NATIVE_SLC5:
        assert "Decided best container to use is" in output, output

    output, used_container = _check_gaudirun(
        ["--allow-containers", "--prefer-container"] + command
    )
    assert used_container, output
    assert "Decided best container to use is" in output, output


def test_setupproject_best():
    command = ["--platform", "best", "Gaudi/v23r0"]
    try:
        output, used_container = _check_gaudirun(["--disallow-containers"] + command)
    except CalledProcessError:
        assert not CAN_RUN_NATIVE_SLC5
    else:
        assert CAN_RUN_NATIVE_SLC5, output
        assert not used_container, output

    output, used_container = _check_gaudirun(
        ["--container", CONTAINER_IMPLEMENTATION] + command
    )
    assert used_container, output

    output, used_container = _check_gaudirun(command)
    assert used_container != CAN_RUN_NATIVE_SLC5, output
    if not CAN_RUN_NATIVE_SLC5:
        assert "Decided best container to use is" in output, output

    output, used_container = _check_gaudirun(
        ["--allow-containers", "--prefer-container"] + command
    )
    assert used_container, output
    assert "Decided best container to use is" in output, output


def test_lbrun_slc6():
    command = ["--platform", "x86_64-slc6-gcc8-opt", "Gaudi/v32r0"]
    try:
        output, used_container = _check_gaudirun(["--disallow-containers"] + command)
    except CalledProcessError:
        assert not CAN_RUN_NATIVE_SLC6
    else:
        assert CAN_RUN_NATIVE_SLC6, output
        assert not used_container, output

    output, used_container = _check_gaudirun(
        ["--container", CONTAINER_IMPLEMENTATION] + command
    )
    assert used_container, output

    output, used_container = _check_gaudirun(command)
    assert used_container != CAN_RUN_NATIVE_SLC6, output
    if not CAN_RUN_NATIVE_SLC6:
        assert "Decided best container to use is" in output, output

    output, used_container = _check_gaudirun(
        ["--allow-containers", "--prefer-container"] + command
    )
    assert used_container, output
    assert "Decided best container to use is" in output, output


def test_lbrun_centos7():
    command = ["--platform", "x86_64-centos7-gcc8-opt", "Gaudi/v32r0"]
    try:
        output, used_container = _check_gaudirun(["--disallow-containers"] + command)
    except CalledProcessError:
        assert not CAN_RUN_NATIVE_CENTOS7
    else:
        assert CAN_RUN_NATIVE_CENTOS7, output
        assert not used_container, output

    output, used_container = _check_gaudirun(
        ["--container", CONTAINER_IMPLEMENTATION] + command
    )
    assert used_container, output

    output, used_container = _check_gaudirun(command)
    assert used_container != CAN_RUN_NATIVE_CENTOS7, output
    if not CAN_RUN_NATIVE_CENTOS7:
        assert "Decided best container to use is" in output, output

    output, used_container = _check_gaudirun(
        ["--allow-containers", "--prefer-container"] + command
    )
    assert used_container, output
    assert "Decided best container to use is" in output, output


def test_lbrun_best():
    command = ["--platform", "best", "Gaudi/v32r0"]
    try:
        output, used_container = _check_gaudirun(["--disallow-containers"] + command)
    except CalledProcessError:
        assert not (CAN_RUN_NATIVE_SLC6 or CAN_RUN_NATIVE_CENTOS7)
    else:
        assert CAN_RUN_NATIVE_SLC6 or CAN_RUN_NATIVE_CENTOS7, output
        assert not used_container, output

    output, used_container = _check_gaudirun(
        ["--container", CONTAINER_IMPLEMENTATION] + command
    )
    assert used_container, output

    output, used_container = _check_gaudirun(command)
    assert used_container != (CAN_RUN_NATIVE_SLC6 or CAN_RUN_NATIVE_CENTOS7), output
    if not (CAN_RUN_NATIVE_SLC6 or CAN_RUN_NATIVE_CENTOS7):
        assert "Decided best container to use is" in output, output

    output, used_container = _check_gaudirun(
        ["--allow-containers", "--prefer-container"] + command
    )
    assert used_container, output
    assert "Decided best container to use is" in output, output


@pytest.mark.parametrize(
    "command,error_message",
    [
        (SETUPPTOJECT_COMMAND, "Cannot find project 'Gaudi'"),
        (
            ["--platform", "best", "Gaudi/v32r0"],
            "current host does not support any of Gaudi/v32r0",
        ),
    ],
)
def test_setupproject_siteroot(command, error_message):
    message = "Using default siteroot of /cvmfs/lhcb.cern.ch/lib"

    output, _ = _check_gaudirun(command)
    assert message in output, output

    output, _ = _check_gaudirun(["--siteroot", "/cvmfs/lhcb.cern.ch/lib"] + command)
    assert message not in output, output

    try:
        _check_gaudirun(["--siteroot", "/cvmfs/lhcb.invalid"] + command)
    except CalledProcessError as e:
        assert error_message in e.output.decode()


_bind_tmp_dir = tempfile.TemporaryDirectory()
bind_tmp_dir = Path(_bind_tmp_dir.name)


@pytest.mark.parametrize("use_env", [True, False])
@pytest.mark.parametrize(
    "bind_specs,test_dir,expect_success",
    [
        ([f"{bind_tmp_dir.parent}:/mounted_path"], "/mounted_path", True),
        (None, "/mounted_path", False),
        ([f"{bind_tmp_dir.parent}:/mounted_path"], "/wrong_path", False),
        (
            [
                f"{bind_tmp_dir.parent}:/mounted_path",
                f"{bind_tmp_dir.parent}:/other_path",
            ],
            "/wrong_path",
            False,
        ),
        (
            [
                f"{bind_tmp_dir.parent}:/mounted_path",
                f"{bind_tmp_dir.parent}:/other_path",
            ],
            "/mounted_path",
            True,
        ),
        (
            [
                f"{bind_tmp_dir.parent}:/mounted_path",
                f"{bind_tmp_dir.parent}:/other_path",
            ],
            "/other_path",
            True,
        ),
    ],
)
def test_bind(use_env, bind_specs, test_dir, expect_success):
    command = SETUPPTOJECT_COMMAND + ["ls", test_dir]
    env = os.environ
    if bind_specs is not None:
        if use_env:
            env = {**env, "LBRUN_BINDS": ";".join(bind_specs)}
        else:
            for bind_spec in bind_specs:
                command = ["--bind", bind_spec] + command
    proc = run(
        ["lb-run"] + command,
        stdout=PIPE,
        stderr=STDOUT,
        text=True,
        check=False,
        env=env,
    )
    if expect_success:
        assert proc.returncode == 0, proc.stdout
        assert "No such file or directory" not in proc.stdout, proc.stdout
        assert bind_tmp_dir.name in proc.stdout, proc.stdout
    else:
        assert proc.returncode != 1, proc.stdout
        assert "No such file or directory" in proc.stdout, proc.stdout



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
@author: Marco Clemencic <marco.clemencic@cern.ch>
"""
import logging
import os
import re
from functools import lru_cache
from itertools import chain
from pathlib import Path

from . import Error, path
from .version import DEFAULT_VERSION, LCGInfoName
from .version import versionKey as _vkey

log = logging.getLogger(__name__)

_PLATFORM_ID_RE = re.compile(
    r"((x86_64|i686)-[a-z]+[0-9]+-[a-z]+[0-9]+-[a-z0-9]+)|"
    r"([a-z]+[0-9]+_[a-z]+[0-9]+_[a-z]+[0-9]+(_dbg)?)"
)

# use uppercase names for case insensitive match
EXTERNAL_PROJECTS = ("ROOT",)

# list of fixed paths for specific projects, used to bypass
# the lookup for nightly builds
forced_paths = {}


def force_paths_from_slot(slot_dir):
    """
    Define forced paths from a nightly build slot configuration in
    the installation directory `slot_dir`.
    """
    import json

    global forced_paths
    try:
        # force all projects in the slot to be used from the slot directory
        slot_config = json.load(open(os.path.join(slot_dir, "slot-config.json")))
        # first collect and then update to avoid partial updates due
        # to malformed configuration
        new_paths = list(
            (project["name"], os.path.join(slot_dir, project["name"]))
            for project in slot_config["projects"]
            if not project.get("disabled")
        )
        forced_paths.update(new_paths)
    except (KeyError, OSError, ValueError):
        # FIXME: it should be `json.JSONDecodeError`, but `ValueError` is good enough for Python 2 and 3
        # ignore invalid/missing/malformed configs
        pass


def platform_sort_key(platform):
    """
    Key function to sort platforms.

    The preferred platform has highest score.
    """
    if "-" not in platform:
        parts = platform.split("_")
        if len(parts) == 3:
            os_id, arch, comp = parts
            opt = "opt"
        elif len(parts) == 4:
            os_id, arch, comp, opt = parts
        else:
            # unknown format
            return tuple()
        arch = {"ia32": "i686", "amd64": "x86_64"}.get(arch, arch)
    elif platform == "<no-platform>":
        # used to identify "platform independent" projects
        # in some test cases it might appear in the list, so
        # better handling it here (as least preferred case)
        return None
    else:
        arch, os_id, comp, opt = platform.split("-")
    os_id = tuple(int(i) for i in os_id if i.isdigit())
    comp = compilerKey(comp)
    return ("0" if opt == "do0" else opt, arch, comp, os_id)


def isPlatformId(s):
    """
    Return True if the string looks like a platform id.

    >>> isPlatformId('x86_64-centos7-gcc64-opt')
    True
    >>> isPlatformId('slc4_ia32_gcc34')
    True
    >>> isPlatformId('include')
    False
    """
    return bool(_PLATFORM_ID_RE.match(s))


def versionKey(x):
    return _vkey(x[0])


def compilerKey(comp):
    family, version, extension = re.match(
        r"^([^\d]+)(\d+)+(\+.+)?$", comp, re.IGNORECASE
    ).groups()
    # Between gcc 3.4 and gcc 6.2, two digits were used for the version
    # Append "0" to all other versions to make the subsequent comparison work
    if not (3 <= int(version[0]) <= 6):
        version += "0"
    version = int(version)
    try:
        family_weight = ["icc", "clang", "gcc"].index(family)
    except ValueError:
        family_weight = -1  # unknown family
    # give highest weight to no extension otherwise sort lexical order
    return family_weight, version, not extension, extension


class NotFoundError(Error):
    """
    Generic error for configuration elements that are not found.
    """

    def __str__(self):
        return f"cannot find {self.args[0]}"


class MissingManifestError(NotFoundError):
    """
    The manifest.xml for a project was not found.
    """


class MissingProjectError(NotFoundError):
    """
    A project was not found.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.name, self.version, self.platform, self.path = args

    def __str__(self):
        return "cannot find project {} {} for {} in {}".format(*self.args)


class MissingDataPackageError(NotFoundError):
    """
    A data package was not found.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.name, self.version, self.path = args

    def __str__(self):
        return "cannot find data package {} {} in {}".format(*self.args)


class InvalidNightlySlotError(NotFoundError):
    """
    A nightly slot build was not found.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.slot, self.build_id, self.path = args

    def __str__(self):
        return "cannot find nightly slot build {}/{}{}.".format(
            self.args[0],
            self.args[1],
            f" in {self.args[2]}" if self.args[2] else "",
        )


def findFile(name, search_path=None):
    """
    Look for a file in the search path.
    """
    log.debug("looking for file %r", name)
    from LbEnv import which

    fn = which(name, search_path or path)
    if fn:
        log.debug("found %r", fn)
    else:
        raise NotFoundError(name)
    return fn


@lru_cache()
def findProject(name, version, platform, allow_empty_version=False):
    """
    Find a Gaudi-based project in the directories specified in the 'path'
    variable.

    @param name: name of the project (case sensitive for local projects)
    @param version: version of the project
    @param platform: binary platform id
    @param allow_empty_version: if True, we allow also the plain project name
                                (without version)

    @return path to the project binary directory

    If name is None, version should be the path to the top directory of the
    project.
    """
    log.debug(
        "findProject(name=%r, version=%r, platform=%r, allow_empty_version=%r)",
        name,
        version,
        platform,
        allow_empty_version,
    )

    if name in forced_paths:
        d = os.path.normpath(os.path.join(forced_paths[name], "InstallArea", platform))
        if os.path.exists(d):
            # name is None when using --path-to-project
            log.debug("using %s for %s", d, name or "main project")
            return d
        else:
            raise MissingProjectError(name, version, platform, forced_paths[name])

    # list of version strings to try
    versions = [version]

    if re.match(r"^\d+\.\d+(\.\d+)*$", version):
        # If we are requested to find a X.Y[.Z[.T]] version (new style CMake config)
        # we have to look also for directories with vXrY[pZ[tN]]
        parts = version.split(".")
        if len(parts) in [2, 3, 4]:
            versions.append("".join(f"{a}{b}" for a, b in zip("vrpt", parts)))
        else:
            pass  # we have a convention only for 2 or 3 parts versions

    # standard project suffixes
    suffixes = []
    for v in versions:
        suffixes.extend(
            [
                f"{name}_{v}",
                os.path.join(name.upper(), f"{name.upper()}_{v}"),
                os.path.join(name.upper(), v),
            ]
        )

    # special case: for the 'latest' version we allow the plain name
    if allow_empty_version:
        suffixes.insert(0, name)

    for d in [
        os.path.normpath(os.path.join(b, s, bindir))
        for b in path
        for s in suffixes
        for bindir in (os.path.join("InstallArea", platform), os.curdir)
    ]:
        log.debug("check %s", d)
        if os.path.exists(d):
            log.debug("OK")
            return d
    else:
        raise MissingProjectError(name, version, platform, path)


def findNightlyDir(slot, build_id, nightly_bases):
    """
    Return the directory of the requested build of a nightly slot, looking in
    the directories listed in nightly_bases.

    If not found raise InvalidNightlySlotError
    """
    # FIXME: we cannot use logging here because this function is called too
    # early in the script

    # log.debug('looking for slot %s %s', slot, build_id)
    slot_bases = [
        Path(nightly_base) / slot_id
        for slot_id in (slot, "lhcb-" + slot, "nightly/" + slot, "nightly/lhcb-" + slot)
        for nightly_base in nightly_bases
    ]
    build_paths = {
        slot_path.name: slot_path
        for slot_base in slot_bases
        for slot_path in (slot_base.iterdir() if slot_base.is_dir() else [])
    }
    if build_id not in build_paths and build_id.lower() == "latest" and build_paths:
        # log.debug("resolving special build id 'latest'")
        build_id = sorted(
            (k for k in build_paths if k.isdigit()), key=int, reverse=True
        )[0]

    if build_id in build_paths:
        # log.debug('found %s', build_paths[build_id])
        return build_paths[build_id].as_posix()

    # log.warning('not found')
    raise InvalidNightlySlotError(slot, build_id, nightly_bases)


def listVersions(name, platform):
    """
    Find all instances of a Gaudi-based project in the directories specified in
    the 'path' variable and return the list of versions found.

    @param name: name of the project (case sensitive for local projects)
    @param platform: binary platform id

    @return generator of pairs (version, fullpath)
    """
    # for special external projects case we delegate to another function
    if name.upper() in EXTERNAL_PROJECTS:
        for entry in listExtVersions(name, platform):
            yield (entry[0], entry[2])
        return
    from .version import isValidVersion

    log.debug("listVersions(name=%r, platform=%r)", name, platform)

    name_u = name.upper()
    prefix = name + "_"
    prefix_u = name_u + "_"
    prefixlen = len(prefix)

    signature = os.path.join("InstallArea", platform, os.curdir)
    if name == "LCG":
        signature = LCGInfoName(platform)
    log.debug("looking for %s in the search path", signature)

    def matches(path):
        # log.debug('testing %s', path)
        return os.path.exists(os.path.join(path, signature)) or platform == "any"

    found_versions = set()
    for p in [dirname for dirname in path if os.path.isdir(dirname)]:
        files = set(os.listdir(p))
        # the plain project name is taken into account as 'default' version
        if DEFAULT_VERSION not in found_versions and name in files:
            fullpath = os.path.join(p, name)
            if matches(fullpath):
                found_versions.add(DEFAULT_VERSION)
                yield (DEFAULT_VERSION, fullpath)

        # versions like 'Project_vXrY'
        for entry in sorted(
            [
                (filename[prefixlen:], os.path.join(p, filename))
                for filename in files
                if (
                    filename.startswith(prefix)
                    and isValidVersion(name, filename[prefixlen:])
                )
            ],
            reverse=True,
            key=versionKey,
        ):
            version, fullpath = entry
            if version not in found_versions and matches(fullpath):
                found_versions.add(version)
                yield entry

        # versions like PROJECT/PROJECT_vXrY
        project_dir = os.path.join(p, name_u)
        if os.path.isdir(project_dir):
            for entry in sorted(
                [
                    (filename[prefixlen:], os.path.join(project_dir, filename))
                    for filename in os.listdir(project_dir)
                    if (
                        filename.startswith(prefix_u)
                        and isValidVersion(name, filename[prefixlen:])
                    )
                ],
                reverse=True,
                key=versionKey,
            ):
                version, fullpath = entry
                if version not in found_versions and matches(fullpath):
                    found_versions.add(version)
                    yield entry


def listPlatforms(name, version, allow_empty_version=False, quiet=False):
    """
    Find a version of a Gaudi-based project in the directories specified in
    the 'path' variable and return the list of platforms available.

    @param name: name of the project (case sensitive for local projects)
    @param version: version string
    @allow_empty_version: search also in directories without the version
    @quiet: if true, do not print a warning if the project is not found

    @return list of platform strings

    If name is None, version should be the path to the top directory of the
    project.
    """
    from os.path import exists, isdir, join, normpath

    # for special external projects case we delegate to another function
    if name and name.upper() in EXTERNAL_PROJECTS:
        return listExtPlatforms(name, version)
    log.debug(
        "listPlatforms(name=%r, version=%r, allow_empty_version=%r)",
        name,
        version,
        allow_empty_version,
    )

    if name:
        # standard project suffixes
        suffixes = [
            f"{name}_{version}",
            join(name.upper(), f"{name.upper()}_{version}"),
            join(name.upper(), version),
        ]
        # special case: for the 'latest' version we allow the plain name
        if allow_empty_version:
            suffixes.insert(0, name)
    else:
        # path to project used, no need for suffixes
        suffixes = [os.curdir]

    platforms = set()
    # if project name is None, version is the path to the top level directory
    # of the project
    for d in [
        normpath(join(b, s)) for b in (path if name else [version]) for s in suffixes
    ]:
        log.debug("check %s", d)
        inst_area = join(d, "InstallArea")
        if isdir(inst_area):
            for platform in os.listdir(inst_area):
                p_dir = join(inst_area, platform)
                if (
                    isdir(p_dir)
                    and isPlatformId(platform)
                    or exists(join(p_dir, "manifest.xml"))
                ):
                    platforms.add(platform)
        elif exists(join(d, "manifest.xml")):
            platforms.add("<no-platform>")

    if not platforms:
        (log.debug if quiet else log.warning)(
            "could not find %s/%s in %r", name, version, path
        )
    return sorted(platforms, key=platform_sort_key, reverse=True)


@lru_cache()
def findDataPackage(name, version):
    """
    Find a data package in the directories specified in the 'path' variable,
    using, optionally, the standard suffixes 'DBASE' and 'PARAM'.
    If version is a pattern, the latest version matching the pattern is
    returned.

    @param name: name of the package with "hat" (case sensitive)
    @param version: glob pattern to filter the version

    @return: the path to the data package
    """
    from fnmatch import fnmatch

    if re.match(r"^\d[\d.]*$", version):
        # this is a X.Y CMake style version, so we map it to vXr*
        # - get the first 3 element max (we only support vXrYpZ in packages)
        parts = version.split(".")[:3]
        # - the first elaments are constrained, but the last is free
        parts[-1] = "*"
        # - join the parts in a string prefixing each of them with the
        #   right positional prefix
        version = "".join(a + b for a, b in zip("vrp", parts))

    suffixes = ["", "DBASE", "PARAM"]
    versions = []
    for p in [os.path.join(b, s, name) for b in path for s in suffixes]:
        if os.path.exists(p):
            lst = os.listdir(p)
            if version in lst:
                # stop searching if we find an exact match
                return os.path.join(p, version)
            versions.extend([(v, p) for v in lst if fnmatch(v, version)])
    if not versions:
        raise MissingDataPackageError(name, version, path)
    # sort the versions found
    versions.sort(key=versionKey, reverse=True)
    v, p = versions[0]
    return os.path.join(p, v)


@lru_cache()
def parseManifest(manifest):
    """
    Extract the list of required projects and data packages from a manifest.xml
    file.

    @param manifest: path to the manifest file
    @return: tuple with ([projects...], [data_packages...]) as (name, version)
             pairs
    """
    from xml.dom.minidom import parse

    m = parse(manifest)

    def _iter(parent, child):
        """
        Iterate over the tags in <parent><child/><child/></parent>.
        """
        for pl in m.getElementsByTagName(parent):
            yield from pl.getElementsByTagName(child)

    # extract the list of used (project, version) from the manifest
    used_projects = [
        (p.attributes["name"].value, p.attributes["version"].value)
        for p in _iter("used_projects", "project")
    ]
    # extract the list of data packages
    data_packages = [
        (p.attributes["name"].value, p.attributes["version"].value)
        for p in _iter("used_data_pkgs", "package")
    ]
    return (used_projects, data_packages)


def getManifests(project, version, platform, allow_empty_version=False):
    """
    Return an iterator over all manifest.xml files from the project to
    the dependencies.
    """
    pdir = findProject(
        project, version, platform, allow_empty_version=allow_empty_version
    )
    manifest = os.path.join(pdir, "manifest.xml")
    if not os.path.exists(manifest):
        raise MissingManifestError(manifest)

    yield manifest

    projects, _ = parseManifest(manifest)
    for p, v in projects:
        for manifest in getManifests(
            p, v, platform, allow_empty_version=allow_empty_version
        ):
            yield manifest


def getEnvXmlPath(project, version, platform, allow_empty_version=False):
    """
    Return the list of directories to be added to the Env XML search path for
    a given project.
    """
    pdir = findProject(
        project, version, platform, allow_empty_version=allow_empty_version
    )
    search_path = [pdir]
    for manifest in getManifests(
        project, version, platform, allow_empty_version=allow_empty_version
    ):
        projects, packages = parseManifest(manifest)
        # add the data package directories
        search_path.extend(findDataPackage(p, v) for p, v in packages)
        # add the project directories
        search_path.extend(
            findProject(p, v, platform, allow_empty_version=False) for p, v in projects
        )

    def _unique(iterable):
        returned = set()
        for i in iterable:
            if i not in returned:
                returned.add(i)
                yield i

    return list(_unique(search_path))


def getProjectsDirs(project, version, platform, allow_empty_version=False):
    """
    Return a dictionary mapping project name to its root directory.
    """
    return {
        getProjectNameVersion(manifest)[0]: os.path.dirname(manifest)
        for manifest in getManifests(
            project, version, platform, allow_empty_version=allow_empty_version
        )
    }


def getPackagesDirs(project, version, platform, allow_empty_version=False):
    """
    Return a dictionary mapping data package name to its root directory.
    """
    out = {}
    for manifest in getManifests(
        project, version, platform, allow_empty_version=allow_empty_version
    ):
        _, packages = parseManifest(manifest)
        out.update((p, findDataPackage(p, v)) for p, v in packages)
    return out


def findLCG(version, platform):
    """
    Return the path to the requested LCG version, found in the search path.
    """
    for p in [
        os.path.join(b, s, n)
        for b in path
        for s in ("", "LCG_%s" % version)
        for n in (f"LCG_{version}_{platform}.txt", LCGInfoName(platform))
    ]:
        if os.path.exists(p):
            return os.path.dirname(p)
    for p in [
        os.path.join(b, "LCGCMT", s)
        for b in path
        for s in ("LCGCMT_%s" % version, "LCGCMT-%s" % version)
    ]:
        if os.path.exists(p):
            return p


def getHepToolsInfo(manifest):
    """
    Extract the hep tools version and platform from a manifest file.
    """
    from xml.dom.minidom import parse

    log.debug("extracting heptools version from %s", manifest)
    try:
        m = parse(manifest)
        heptools = m.getElementsByTagName("heptools")[0]
        version = heptools.getElementsByTagName("version")[0].firstChild.nodeValue
        platform_entries = heptools.getElementsByTagName("lcg_platform")
        if platform_entries:
            platform = platform_entries[0].firstChild.nodeValue
        else:
            platform = heptools.getElementsByTagName("binary_tag")[
                0
            ].firstChild.nodeValue
        return version, platform
    except (IndexError, AttributeError) as exc:
        # cannot extract heptools version and platform
        raise NotFoundError("heptools info: %s" % exc)


def getLCGRelocation(manifest):
    from os.path import basename, dirname, join, pardir

    try:
        version, platform = getHepToolsInfo(manifest)
    except NotFoundError as exc:
        log.debug(str(exc))
        return {}

    log.debug("looking for LCG %s %s", version, platform)
    lcg_path = findLCG(version, platform)
    if not lcg_path:
        log.debug("cannot find LCG path")
        return {}
    # FIXME: xenv has got problems with unicode filenames
    lcg_path = str(lcg_path)
    log.debug("found LCG at %s", lcg_path)

    if basename(dirname(lcg_path)) == "LCGCMT":
        # old style
        lcg_root = dirname(dirname(lcg_path))
        return {
            "LCG_releases": lcg_root,
            "LCG_external": (
                lcg_root
                if lcg_root.endswith("external")
                else join(lcg_root, pardir, pardir, "external")
            ),
        }
    else:
        # new style
        return {
            "LCG_releases": lcg_path,
            "LCG_external": lcg_path,
            "LCG_releases_base": (
                dirname(lcg_path) if lcg_path.endswith("LCG_%s" % version) else lcg_path
            ),
        }


def getProjectNameVersion(manifest):
    """
    Get project name and version from a manifest.xml.
    """
    from xml.dom.minidom import parse

    log.debug("extracting project name from %s", manifest)
    name, version = None, None
    try:
        m = parse(manifest)
        project = m.getElementsByTagName("project")[0]
        name = project.getAttribute("name")
        version = project.getAttribute("version")
    except (IndexError, AttributeError) as exc:
        log.debug("cannot extract project name or version: %s", exc)
    return name, version


def listExtVersions(ext, platform):
    """
    @return generator of tuples (ext_version, LCG_version, LCG_path)
    """
    ext = ext.upper()  # case insensitive check
    found_versions = set()
    for LCG_version, LCG_path in listVersions("LCG", platform):
        for l in open(os.path.join(LCG_path, LCGInfoName(platform))):
            l = l.split(";")
            if l[0].strip().upper() == ext:
                v = l[2].strip()
                if v not in found_versions:
                    found_versions.add(v)
                    yield (l[2].strip(), LCG_version, LCG_path)
                break  # go to next LCG version


def listExtPlatforms(ext, version):
    """
    @return list of available platforms for a version of an extrenal project
    """
    import glob

    log.debug("listExtPlatforms(ext=%r, version=%r)", ext, version)

    ext = ext.upper()  # case insensitive check

    info_glob = LCGInfoName("*")
    # transform the glob in a capturing regexp
    platform_exp = re.compile(re.escape(info_glob).replace("\\*", "(.*)"))

    platforms = set()
    for p in [os.path.join(b, s, info_glob) for b in path for s in ("", "LCG_*")]:
        for f in glob.glob(p):
            # it must match if the glob matched
            platform = platform_exp.search(f).group(1)
            # to be sure, check if the external is in the list with the right
            # version
            for l in (l.split(";") for l in open(f)):
                if l[0].strip().upper() == ext and l[2].strip() == version:
                    platforms.add(platform)
                    break
    return sorted(platforms, key=platform_sort_key, reverse=True)


def findLCGForExt(ext, version, platform):
    """
    Return the (highest) version of LCG containing the required version of
    an external project.
    """
    for ext_vers, LCG_vers, _path in listExtVersions(ext, platform):
        if ext_vers == version:
            return LCG_vers
    raise MissingProjectError(ext, version, platform, path)


PREFERRED_PLATFORM = os.environ.get("BINARY_TAG") or os.environ.get("CMTCONFIG") or None


def _findManifest(path, platform_hint=PREFERRED_PLATFORM):
    """
    Find a manifest file of the project rooted at 'path'.
    """
    from os.path import exists, isdir, join

    # try the InstallArea level for the current platform
    inst = join(path, "InstallArea")

    def candidates():
        if platform_hint:
            yield join(inst, platform_hint, "manifest.xml")
        if isdir(inst):
            for p in os.listdir(inst):
                yield join(inst, p, "manifest.xml")
        yield join(path, "manifest.xml")

    for c in candidates():
        if exists(c):
            return c
    return None


def walkProjectDeps(project, version, platform="any", ignore_missing=False):
    """
    Return a tuple ((project, version), rootdir, dependencies) for the
    requested project and all its dependencies.

    The dependencies are returned as a list of (project, version) and the list
    can be altered to control which dependency to follow and which not.

    Note that when walking through the dependencies a project may appear more
    than once (for diamond dependency graphs).
    """
    if project.lower() in ("lcg", "heptools", "lcgcmt"):
        return
    try:
        root = findProject(project, version, platform, allow_empty_version=False)
    except MissingProjectError as err:
        if ignore_missing:
            log.warning(str(err))
            return
        else:
            raise
    manifest = _findManifest(root)
    if manifest:
        deps = [(p, v) for p, v in parseManifest(manifest)[0]]
        try:
            deps.append(("lcg", getHepToolsInfo(manifest)[0]))
        except NotFoundError:
            pass
    else:
        deps = []
    yield (project, version), root, deps
    for project, version in deps:
        yield from walkProjectDeps(project, version, platform, ignore_missing)


def getLHCbGrid(platform):
    """
    Find the best LHCbGrid version and platform for the requested platform.
    """
    from LbEnv.ProjectEnv.version import expandVersionAlias

    for version in [DEFAULT_VERSION, "latest"]:
        version = expandVersionAlias("LHCbGrid", "latest", "any")
        platforms = listPlatforms("LHCbGrid", version, quiet=True)
        if platforms:
            break
    else:
        raise NotFoundError("LHCbGrid")
    if platform not in platforms:
        log.warning("platform %s not available for LHCbGrid", platform)
        os_id = platform.split("-")[1]
        for platform in platforms:
            if platform.split("-")[1] == os_id:
                log.warning("using %s instead", platform)
                break
    return version, platform



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
import sys

import importlib_resources
from importlib_metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "unknown"

DEFAULT_SITEROOT = "/cvmfs/lhcb.cern.ch/lib"


def computeDefaultSiteRoot():
    if "MYSITEROOT" in os.environ:
        mysiteroot = os.environ["MYSITEROOT"]
    elif "VO_LHCB_SW_DIR" in os.environ:
        mysiteroot = os.path.join(os.environ["VO_LHCB_SW_DIR"], "lib")
    elif "VIRTUAL_ENV" in os.environ:
        mysiteroot = os.environ["VIRTUAL_ENV"]
        # the venv is usually in #MYSITEROOT/var/..., so we try to deduce it
        head, tail = mysiteroot, "none"
        while tail and tail != "var":
            head, tail = os.path.split(head)
        if tail:
            # 'var' was found in the path, so the parent directory is the root
            mysiteroot = head
    elif os.path.isdir(DEFAULT_SITEROOT):
        mysiteroot = DEFAULT_SITEROOT
    else:
        sys.stderr.write("error: not valid siteroot provided\n")
        sys.exit(1)
    return mysiteroot


def which(name, path=None):
    """
    Locate a file in the path.
    """
    if path is None:
        path = os.environ.get("PATH", "")
    if isinstance(path, str):
        path = path.split(os.path.pathsep)
    for p in path:
        fp = os.path.join(p, name)
        if os.path.exists(fp):
            return fp
    return None


def resource_string(name):
    """
    Helper to get data stored with the package.
    """
    ref = importlib_resources.files(__name__).joinpath("data/" + name)
    return ref.read_text()


def getProjectNames():
    """
    Return an iterator over the known project names.
    """
    for line in resource_string("projects.txt").splitlines():
        # remove comments and whitespaces
        project = line.split("#", 1)[0].strip()
        if project:
            yield project


def getPackageNames():
    """
    Return an iterator over the known project names.
    """
    for line in resource_string("packages.txt").splitlines():
        # remove comments and whitespaces
        package = line.split("#", 1)[0].strip()
        if package:
            yield package


_PROJECT_NAMES = None


def fixProjectCase(project):
    """
    Convert a project name to its canonical case, if known, otherwise return
    the string unchanged.

    >>> fixProjectCase('gaudi')
    'Gaudi'
    >>> fixProjectCase('DAvinci')
    'DaVinci'
    >>> fixProjectCase('UnKnown')
    'UnKnown'
    """
    global _PROJECT_NAMES
    if _PROJECT_NAMES is None:
        _PROJECT_NAMES = {name.lower(): name for name in getProjectNames()}
    return _PROJECT_NAMES.get(project.lower(), project)



from LbEnv import getPackageNames, getProjectNames


def test_getProjectNames():
    names = list(getProjectNames())
    assert "Gaudi" in names
    assert "DaVinci" in names
    assert "Gauss" in names
    assert "AppConfig" not in names
    assert "WG/SemilepConfig" not in names


def test_getPackageNames():
    names = list(getPackageNames())
    assert "Gaudi" not in names
    assert "DaVinci" not in names
    assert "Gauss" not in names
    assert "AppConfig" in names
    assert "WG/SemilepConfig" in names



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Functions to be able to use the old SetupProject (CMT) from within
LbEnv.ProjectEnv.
"""


__author__ = "Marco Clemencic <marco.clemencic@cern.ch>"

import os
import re
import sys
import time
from fnmatch import fnmatch
from stat import S_ISDIR
from tempfile import mkstemp
from xml.sax import ContentHandler, parse

from LbEnv import computeDefaultSiteRoot, fixProjectCase
from LbEnv.Utils.Temporary import TempDir

__version__ = "legacy"

try:
    from LbScriptsPolicy.PathStripper import StripPath
except ImportError:
    # Failover solution
    StripPath = lambda x: x

import logging

# Import all levels constants for consistency with VERBOSE and ALWAYS defined below
from logging import DEBUG, ERROR, FATAL, INFO, WARNING
from subprocess import PIPE, STDOUT, Popen

# Extend the standard logging with one level that sits in the middle between INFO and DEBUG
VERBOSE = (INFO + DEBUG) / 2
logging.addLevelName(VERBOSE, "VERBOSE")
ALWAYS = FATAL + 10
logging.addLevelName(ALWAYS, "ALWAYS")


## Specialization of the default Python Logger class to use two more levels (a la LHCb)
class ExtendedLogger(logging.Logger):
    def verbose(self, msg, *args, **kwargs):
        """
        Cloned from the default member functions in logging.Logger.
        """
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, msg, args, **kwargs)

    def always(self, msg, *args, **kwargs):
        """
        Cloned from the default member functions in logging.Logger.
        """
        if self.isEnabledFor(ALWAYS):
            self._log(ALWAYS, msg, args, **kwargs)


logging.setLoggerClass(ExtendedLogger)
log = logging.getLogger(os.path.basename(__file__))

########################################################################
# Useful constants
########################################################################
lhcb_style_version = re.compile(r"v([0-9]+)r([0-9]+)(?:p([0-9]+))?")
lcg_style_version = re.compile(r"([0-9]+)([a-zA-Z_0-9]*)$")

# LHCb projects without container
# LCGCMT is not an LHCb project but we know about it
nocontainer_project_names = ["DABASE", "PARAM", "LCGCMT", "LCG", "ROOT"]

# List of pairs (project,[packages]) to automatically select for override
# The project are prepended to the list of overriding packages and
# the packages are appended to the list of used packages
auto_override_projects = [
    ("ExtraPackages", []),
    # ("LHCbGrid", ["LHCbGridSys"]), # enabled with --use-grid
]

# List of abbreviations for weekdays
days = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

nightly_bases = []

########################################################################
# Utility classes
########################################################################
if sys.platform.startswith("win"):
    _fixKey = lambda key: key.upper()
else:
    _fixKey = lambda key: key


class TemporaryEnvironment:
    """
    Class to changes the environment temporarily.
    """

    def __init__(self, orig=os.environ, keep_same=False):
        """
        Create a temporary environment on top of the one specified
        (it can be another TemporaryEnvironment instance).
        """
        # print "New environment"
        self.old_values = {}
        self.env = orig
        self._keep_same = keep_same

    def __setitem__(self, key, value):
        """
        Set an environment variable recording the previous value.
        """
        key = _fixKey(key)
        if key not in self.old_values:
            if key in self.env:
                if not self._keep_same or self.env[key] != value:
                    self.old_values[key] = self.env[key]
            else:
                self.old_values[key] = None
        self.env[key] = value

    def __getitem__(self, key):
        """
        Get an environment variable.
        Needed to provide the same interface as os.environ.
        """
        key = _fixKey(key)
        return self.env[key]

    def __delitem__(self, key):
        """
        Unset an environment variable.
        Needed to provide the same interface as os.environ.
        """
        key = _fixKey(key)
        if key not in self.env:
            raise KeyError(key)
        # Record that we unset a variable only if it was not set before
        if (key in self.old_values) and (self.old_values[key] is None):
            del self.old_values[key]
        else:
            self.old_values[key] = self.env[key]
        del self.env[key]

    def __len__(self):
        """
        Return the size of the internal dictionary.
        Needed for the conversion to boolean.
        """
        return len(self.env)

    def keys(self):
        """
        Return the list of defined environment variables.
        Needed to provide the same interface as os.environ.
        """
        return self.env.keys()

    def has_key(self, key):
        """
        return True if the key is present
        """
        key = _fixKey(key)
        return key in self.env

    def items(self):
        """
        Return the list of (name,value) pairs for the defined environment variables.
        Needed to provide the same interface as os.environ.
        """
        return self.env.items()

    def __contains__(self, key):
        """
        Operator 'in'.
        Needed to provide the same interface as os.environ.
        """
        key = _fixKey(key)
        return key in self.env

    def restore(self):
        """
        Revert all the changes done to the original environment.
        """
        for key, value in self.old_values.items():
            if value is None:
                del self.env[key]
            else:
                self.env[key] = value
        self.old_values = {}

    def __del__(self):
        """
        Revert the changes on destruction.
        """
        # print "Restoring the environment"
        self.restore()

    def get(self, key, default=None):
        """
        Implementation of the standard get method of a dictionary: return the
        value associated to "key" if present, otherwise return the default.
        """
        key = _fixKey(key)
        return self.env.get(key, default)

    def commit(self):
        """
        Forget the old values for the changes done so far (avoids that the
        changes are rolled-back when the instance goes out of scope).
        """
        self.old_values = {}

    def gen_script(self, shell_type):
        """
        Generate a shell script to reproduce the changes in the environment.
        """
        var_re = re.compile(r"(?:\$(\w+|\{[^}]*\})|%(\w+)%)")
        shells = ["csh", "sh", "bat"]
        if shell_type not in shells:
            raise RuntimeError(
                f"Shell type '{shell_type}' unknown. Available: {shells}"
            )
        out = ""
        old_vars = list(self.old_values.keys())
        done_vars = []
        while old_vars:
            key = old_vars.pop(0)
            if key not in self.env:
                # unset variable
                if shell_type == "csh":
                    out += "unsetenv %s\n" % key
                elif shell_type == "sh":
                    out += "unset %s\n" % key
                elif shell_type == "bat":
                    out += "set %s=\n" % key
                done_vars.append(key)
            else:
                val = self.env[key]
                can_be_added = True
                if "$" in val or "%" in val:
                    ms = var_re.findall(val)
                    for group1, group2 in ms:
                        ref_var = group1 or group2
                        if ref_var.startswith("{") and ref_var.endswith("}"):
                            ref_var = ref_var[1:-1]
                        if ref_var not in done_vars and ref_var in old_vars:
                            can_be_added = False
                            break
                if can_be_added:
                    # set variable
                    if shell_type == "csh":
                        out += f'setenv {key} "{self.env[key]}"\n'
                    elif shell_type == "sh":
                        out += f'export {key}="{self.env[key]}"\n'
                    elif shell_type == "bat":
                        out += f"set {key}={self.env[key]}\n"
                    done_vars.append(key)
                else:
                    old_vars.append(key)
        return out

    def __iter__(self):
        """
        Forward the iteration to the internal dictionary.
        """
        return self.env.__iter__()


def _sync_dicts(src, dest):
    # remove undefined keys
    keys = list(dest.keys())
    for k in keys:
        if k not in src:
            del dest[k]
    # set the variables
    for k in src:
        dest[k] = src[k]


def isProject(path, ignore_not_ready=False, cvmfs=True):
    # It is a project directory if it contains cmt/project.cmt and it doesn't
    # contain NOT_READY (unless we are told to ignore that flag).
    # NOT_READY -> projects that are being released
    if cvmfs:
        # On CVMFS it's too heavy to look for the project.cmt file, so we rely on
        # the naming convention (path/to/PROJECT/PROJECT_vXrY).
        try:
            # take the last two elements of the path
            p, v = path.rsplit(os.path.sep, 2)[-2:]
            # check that they are like PROJECT and PROJECT_vXrY
            if v.startswith(p + "_"):
                v = v[len(p) + 1 :]
                return lhcb_style_version.match(v) or lcg_style_version.match(v)
            else:
                # in case of projects without version, 'v' is the name of the project
                return v in ("DBASE", "PARAM")
        except:
            return False
    # on other filesystems we use a better check
    return os.path.isfile(os.path.join(path, "cmt", "project.cmt")) and (
        ignore_not_ready or not os.path.exists(os.path.join(path, "NOT_READY"))
    )


def _extract_version(project, version):
    # By default, the version is the name of the version directory
    v = version
    # if the version directory starts with the project name, we use the
    # exceeding part (GAUDI_v20r0 -> v20r0)
    if version.startswith(project):
        if len(version) > len(project):  # the version directory could be == project
            v = version[len(project) :].lstrip("_")
            if not v:  # handle the case "Gaudi_" considering '_' as version
                v = version[len(project) :]
    return v


def FindProjectVersions(
    project, search_path, user_area=None, ignore_not_ready=False, requested_version=None
):
    """Given a project name, discovers all the matching project/versions using
    the provided user_area and the the search_path (a list of directories).
    Returns a list of tuples with (name, version, realname, basepath) where
      'name'     is the canonical name of the project (e.g. Gaudi, DaVinci)
      'version'  is the canonical version string (e.g. v2r3)
      'realname' is the name of the root directory of the project (e.g GAUDI/GAUDI_v21r6,
                 LHCb_v28r3, DBASE)
      'basepath' is the directory in which 'realname' is based.
    If there is no version the second entry is None.
    """
    versions = []  # container of the results
    # Look into user_area for projects (without version directory)
    if user_area and os.path.isdir(user_area):
        user_projects = [
            d
            for d in os.listdir(user_area)
            if isProject(os.path.join(user_area, d), ignore_not_ready, False)
        ]  # User area never on CVMFS
        # look for projects with names starting with <project>
        candidates = [d for d in user_projects if d.startswith(project)]
        # If there is no user project with the given name, try to use the
        # correct case for the project
        if not candidates:
            prj = fixProjectCase(project)
            candidates = [d for d in user_projects if d.startswith(prj)]
        # extract 'fake' version informations
        for c in candidates:
            if c == project:
                v = None
            else:
                v = _extract_version(project, c)
            versions.append((project, v, c, user_area))

    candidates = []
    # find all project directories in the searchpath
    PROJECT = project.upper()
    # search_path = os.environ["CMTPROJECTPATH"].split(os.path.pathsep)

    # special cases for the version, if explicitly requested
    special_project_dirs = []
    if requested_version:
        # user project or nightlies like version
        special_project_dirs.append(f"{project}_{requested_version}")

        prj = fixProjectCase(project)
        if prj != project:
            # same as above, but with canonical project name (if applicable)
            special_project_dirs.append(f"{prj}_{requested_version}")

        # release style project name+version
        special_project_dirs.append(
            os.path.join(PROJECT, f"{PROJECT}_{requested_version}")
        )

    for d in search_path:
        versions.extend(
            (project, requested_version, v, d)
            for v in special_project_dirs
            if os.path.isdir(os.path.join(d, v))
        )
        p = os.path.join(d, PROJECT)
        if os.path.isdir(p):
            if isProject(p, ignore_not_ready, cvmfs=True):
                # project without version
                versions.append((project, None, PROJECT, d))
            else:
                # project with versions
                versions += [
                    (project, _extract_version(PROJECT, v), os.path.join(PROJECT, v), d)
                    for v in os.listdir(p)
                    if isProject(os.path.join(p, v), ignore_not_ready, cvmfs=True)
                ]
    return versions


def FindLCGForROOT(version, versions):
    """
    Return the version of LCGCMT for the required version of ROOT.

    @param versions: list of tuples like the one returned by FindProjectVersions
    """
    from os.path import exists, join

    root_version = re.compile(r'^\s*macro\s+ROOT_config_version\s+"([^"]*)"')
    for req, vers in [
        (join(base, name, "LCG_Configuration", "cmt", "requirements"), vers)
        for _, vers, name, base in versions
    ]:
        if exists(req):
            for l in open(req):
                m = root_version.match(l)
                if m:
                    if m.group(1) == version:
                        return vers  # got it
                    else:
                        break  # we found the declaration of ROOT, but we do not like it
    return None


def SortVersions(versions, reverse=False):
    """Give a list of version numbers, return a list with only unique elements
    ordered by version.
    """
    # remove duplicates
    tmp_list = uniq(versions)
    sortable_list = []
    for v in tmp_list:
        if v is None:
            sortable_list.append(((0, 0, v), v))
        else:
            m = lhcb_style_version.match(v)
            if m:
                a, b, c = m.groups()
                a = int(a)
                b = int(b)
                if c:
                    c = int(c)
                else:
                    c = 0
                sortable_list.append(((a, b, c), v))
            else:
                m = lcg_style_version.match(v)
                if m:
                    a, b = m.groups()
                    a = int(a)
                    sortable_list.append(((0, a, b), v))
                elif v == "HEAD":
                    sortable_list.append(((10000, 0, 0), v))
                else:
                    sortable_list.append(((0, 0, v), v))
    sortable_list.sort(reverse=reverse)
    return [v[1] for v in sortable_list]


def SortVersionTuples(versions, reverse=False):
    """
    Similar to SortVersions, but take the list of versions returned by FindVersions.
    """
    vers_map = {t[1]: t for t in versions[::-1]}
    return [vers_map[k] for k in SortVersions(vers_map, reverse)]


def LatestVersion(versions):
    """Extract the latest version from a list of version tuples
    (output of FindProjectVersions)."""
    return SortVersions([v[1] for v in versions])[-1]


def VersionMatch(version, pattern):
    """Compare a version string with a pattern. The pattern can be
    the exact match, a glob pattern or None (in which case matches a None version).
    """
    return version == pattern or (
        (pattern is not None) and (version is not None) and fnmatch(version, pattern)
    )


def _GetVersionTuple(pattern, versions):
    """Extract the version tuple corresponding to version 'pattern'."""
    if not versions:
        # No need to go on if the versions list is empty
        return None
    # Extract the list of version strings (in the correct order)
    version_strings = SortVersions([v[1] for v in versions], reverse=True)
    # Look for a match in the list of versions
    found = False
    match = None
    for v in version_strings:
        if VersionMatch(v, pattern):  # compare the version with the pattern
            found = True
            match = v
            break  # exit the loop at the first match
    if not found and pattern is None:  # fall back solution
        # Normal behavior
        match = version_strings[0]  # latest version

    # Now that we have a string (and not a pattern), we can extract the tuple
    for vers_tuple in versions:
        if match == vers_tuple[1]:
            return vers_tuple
    # Nothing found
    return None


class ProjectInfo:
    """
    Class to get project related infos.
    """

    # @todo: use LHCB_config.py instead of scanning directories
    searchpath = []
    # Projects for which the check for the platform does not make sense.
    no_platform_projects = [
        "Compat",
        "LbScripts",
        "LCGCMT",
        "ExtraPackages",
        "LHCbDirac",
        "LHCbGrid",
    ] + nocontainer_project_names

    def __init__(self, project, version, realName, path):
        self.name = fixProjectCase(project)
        self.version = version
        self.realName = realName
        self.path = path
        self.project_dir = os.path.join(self.path, self.realName)

        # discover project policy
        if self.version:
            self._projectenv_cmt_dir = os.path.join(
                self.path, "%sEnv" % self.name, self.version, "cmt"
            )
            if os.path.isdir(self._projectenv_cmt_dir):
                self.policy = "old"
            else:
                self._projectenv_cmt_dir = None
                self.policy = "new"
        else:
            # Projects without version must be considered "new" style
            self._projectenv_cmt_dir = None
            self.policy = "new"

        self.sys = self._container()
        if self.version and self.sys:
            self.syscmtfullpath = os.path.join(
                self.project_dir, self.sys, self.version, "cmt"
            )
        else:
            self.syscmtfullpath = None

    def _container(self):
        """
        Return the name of the package referencing all the packages belonging to a project.
        """
        container = None
        project_cmt = os.path.join(self.path, self.realName, "cmt", "project.cmt")
        if os.path.exists(project_cmt):
            for l in open(project_cmt):
                tokens = l.split()
                if tokens and tokens[0] == "container":
                    container = tokens[1]
                    break
        if container is None:
            if self.name == "Gaudi":
                if self.policy == "new":
                    container = self.name + "Release"
                else:
                    container = self.name
            elif self.name in nocontainer_project_names:  # Project without container
                container = None
            else:
                container = self.name + "Sys"
        return container

    def __str__(self):
        if self.version:
            return f"{self.name} {self.version} from {self.project_dir}"
        else:
            return f"{self.name} from {self.project_dir}"

    def supportsPlatform(self, platform, user_area=None):
        """
        Check if the project supports the requested platform (i.e. it contains
        the correct platform directory in the InstallArea).
        """
        return (
            self.path == user_area
            or self.name in self.no_platform_projects
            or os.path.isdir(os.path.join(self.project_dir, "InstallArea", platform))
        )


def _defaultSearchPath(env=None):
    search_path = []
    if env is None:
        env = os.environ
    for v in ["CMTPROJECTPATH", "CMAKE_PREFIX_PATH", "LHCBPROJECTPATH"]:
        if v in env:
            search_path += env[v].split(os.pathsep)
    return search_path


def makeProjectInfo(
    project=None,
    version=None,
    versions=None,
    search_path=None,
    user_area=None,
    env=None,
    ignore_not_ready=False,
):
    # actual body
    if versions is None:
        if not project:
            raise TypeError("makeProjectInfo() requires either 'project' or 'versions'")
        if not search_path:  # default search path
            search_path = _defaultSearchPath(env)
            # raise TypeError("makeProjectInfo() requires 'search_path' if 'versions' is not specified")
        versions = FindProjectVersions(
            project, search_path, user_area, ignore_not_ready
        )
    vers_tuple = _GetVersionTuple(version, versions)
    if not vers_tuple:
        return None
    return ProjectInfo(*vers_tuple)


if sys.platform.startswith("win"):
    # On Windows, Popen needs a list (when using shell=True)
    _prepare_cmt_cmd = lambda cmd, args: ["cmt", cmd] + args
else:
    # on Unix, we need a string
    _prepare_cmt_cmd = lambda cmd, args: " ".join(
        ["cmt", cmd] + ['"%s"' % s for s in args]
    )


def cmt(command, args=[], environment=None, cwd=None):
    if type(args) is str:
        args = [args]
    cmd = _prepare_cmt_cmd(command, args)
    if not cwd:
        cwd = os.getcwd()
    if not environment:
        environment = os.environ
    return (
        Popen(cmd, shell=True, env=environment, cwd=cwd, stdout=PIPE, stderr=STDOUT)
        .communicate()[0]
        .decode()
    )


class GetNightlyCMTPROJECTPATH(ContentHandler):
    """SAX content handler to extract the CMTPROJECTPATH from lcg nightly build
    configuration for a given slot/day.
    Usage:
    from xml.sax import parse
    getter = GetNightlyCMTPROJECTPATH("lhcb2","Mon")
    parse(filename, getter)
    print(getter.value())
    """

    def __init__(self, slot, day=None):
        ContentHandler.__init__(self)
        self.slot = slot
        self._path = []
        self._in_slot = False
        self._in_cmtpp = False
        if day in (None, "Today"):
            self._day = time.localtime()[6]  # today
        elif day in days:
            self._day = days.index(day)
        elif day in range(7):
            self._day = day
        else:
            raise TypeError("What do you mean with day = %r?" % day)

    def startElement(self, name, attrs):
        """SAX implementation."""
        if name == "slot":
            if attrs["name"] == self.slot:
                self._in_slot = True
        elif name == "cmtprojectpath":
            self._in_cmtpp = True
        elif name == "path":
            if self._in_slot and self._in_cmtpp:
                self._path.append(attrs["value"])

    def endElement(self, name):
        """SAX implementation."""
        if name == "slot":
            self._in_slot = False
        elif name == "cmtprojectpath":
            self._in_cmtpp = False

    def dirs(self):
        """Return the list of entries in the CMTPROJECTPATH."""
        today = days[self._day]
        yesterday = days[self._day - 1]
        d = [
            s.replace("%DAY%", today).replace("%YESTERDAY%", yesterday)
            for s in self._path
        ]
        return d

    def value(self):
        """Return the CMTPROJECTPATH."""
        return os.path.pathsep.join(self.dirs())


def getNightlyCMTPROJECTPATH(path, slot, day):
    """Simple wrapper around GetNightlyCMTPROJECTPATH.
    Returns the list of entries in the CMTPROJECTPATH of the slot of the nightly
    build.
    """
    getter = GetNightlyCMTPROJECTPATH(slot, day)
    parse(path, getter)
    return getter.dirs()


def smartExpandVarsPath(path, env=None):
    """
    Expand the environment variables in a path. If the variable in the entry of
    the path is a path itself, one entry is added for each element of the
    variable. (The code is more or less inspired by posixpath.py)
    Note: replaces only the first variable in each entry.
    """
    if env is None:
        env = os.environ
    entries = path.split(os.pathsep)
    new_entries = []
    # matches:
    #   $varname
    #   ${varname}
    #   %varname%
    var_re = re.compile(r"(?:\$(\w+|\{[^}]*\})|%(\w+)%)")
    for e in entries:
        if "$" not in e and "%" not in e:
            new_entries.append(e)
        else:
            match = var_re.search(e)
            if match:
                name = match.group(1) or match.group(2)
                if name.startswith("{") and name.endswith("}"):
                    name = name[1:-1]
                if name not in env:
                    new_entries.append(e)
                else:
                    a, b = match.span(0)
                    ea, eb = e[:a], e[b:]
                    for ve in env[name].split(os.pathsep):
                        new_entries.append(ea + ve + eb)
    return os.pathsep.join(new_entries)


########################################################################
# Utility functions
########################################################################


def uniq(iterable):
    """Returns a list of unique elements in the passed iterable."""
    result = []
    for i in iterable:
        if i not in result:
            result.append(i)
    return result


def removeall(path):
    """
    Recursively remove directories and files.
    """
    log.debug("----- removeall(%s) -----", path)
    if S_ISDIR(os.stat(path).st_mode):
        lst = os.listdir(path)
        for p in lst:
            removeall(os.path.join(path, p))
        os.rmdir(path)
    else:
        os.remove(path)


def _get_cmt_user_context(env=None):
    log.debug("----- get_cmt_user_context() -----")
    if env is None:
        env = os.environ
    if "CMTPATH" not in env:
        # This function needs CMTPATH, which means that it does not work with CMTPROJECTPATH
        log.debug("----- get_cmt_user_context(): no CMTPATH -----")
        return None  # For the moment we ignore the CMTUSERCONTEXT

    # check if we can use CMTUSERCONTEXT
    context_path = None
    context_validity = None
    current_lcgcmt = None

    # Find the context validity
    if env["LHCb_release_area"]:
        context_path = os.path.join(env["LHCb_release_area"], "context")
        context_file = os.path.join(context_path, "requirements")

        if os.access(context_file, os.F_OK | os.R_OK):
            # find the magic comment
            for l in open(context_file):
                m = re.match("^#.*created.*from.*LCGCMT_(([0-9]+)([a-z]?)|HEAD)", l)
                if m:
                    context_validity = m.groups()
                    break

    if not context_validity:
        log.debug("----- get_cmt_user_context(): context not found -----")
        return None
    else:
        log.debug(
            "----- get_cmt_user_context(): context if from LCGCMT %s -----",
            context_validity[0],
        )

    # we have to compare the version with the current one
    for l in env["CMTPATH"].split(os.pathsep):
        m = re.match("^.*LCGCMT_(([0-9]+)([a-z]?)|HEAD)", l)
        if m:
            current_lcgcmt = m.groups()
            break
    if current_lcgcmt:
        log.debug(
            "----- get_cmt_user_context(): Default LCGCMT is version %s -----",
            current_lcgcmt[0],
        )

    if not current_lcgcmt:
        # for no lcgcmt in the path, assume we need the context
        log.debug(
            "----- get_cmt_user_context(): LCGCMT not found, use %s -----", context_path
        )
        return context_path

    # at this point we have both context and current: comparison time
    if context_validity[0] == current_lcgcmt[0] or current_lcgcmt[0] == "HEAD":
        # the current is good
        log.debug("----- get_cmt_user_context(): current LCGCMT is good -----")
        return None

    if context_validity[0] == "HEAD":
        # the current cannot be better
        log.debug("----- get_cmt_user_context(): use %s -----", context_path)
        return context_path

    if int(context_validity[1]) > int(current_lcgcmt[1]):
        # context is better
        log.debug("----- get_cmt_user_context(): use %s -----", context_path)
        return context_path
    elif (context_validity[1] == current_lcgcmt[1]) and (
        context_validity[2] > current_lcgcmt[2]
    ):
        # like in 46b > 46a.. context is still better
        log.debug("----- get_cmt_user_context(): use %s -----", context_path)
        return context_path
    else:
        # it seems that the current one is better
        log.debug("----- get_cmt_user_context(): do not use the context -----")
        return None


def shParser(script, env):
    leftover = []
    for l in script.splitlines():
        if l.startswith("unset"):
            var = l.split()[-1]
            if var in env:
                del env[var]
        elif "export " in l:
            l = l.split(";")[0]
            var, value = l.split("=", 1)
            value = os.path.expandvars(value.strip('"'))  # unix style vars
            env[var] = value
        else:
            leftover.append(l)
    if leftover:
        leftover.append("")  # an empty line to have the final newline
    return "\n".join(leftover)


def cshParser(script, env):
    leftover = []
    for l in script.splitlines():
        if l.startswith("unsetenv"):
            var = l.split()[-1]
            if var in env:
                del env[var]
        elif l.startswith("setenv"):
            dummy, var, value = l.split(None, 2)
            value = os.path.expandvars(value.strip('"'))  # unix style vars
            env[var] = value
        else:
            leftover.append(l)
    if leftover:
        leftover.append("")  # an empty line to have the final newline
    return "\n".join(leftover)


def batParser(script, env):
    win_var = re.compile(r"%([^0-9][^%]*)%")
    leftover = []
    for l in script.splitlines():
        if l.startswith("set"):
            l = l.split(None, 1)[1]
            var, value = l.split("=", 1)
            value = os.path.expandvars(value)  # unix style vars
            value = win_var.sub(r"%(\1)s", value) % env  # win style vars
            env[var] = value
        else:
            leftover.append(l)
    if leftover:
        leftover.append("")  # an empty line to have the final newline
    return "\n".join(leftover)


ShellParser = {"sh": shParser, "csh": cshParser, "bat": batParser}


########################################################################
# Main class
########################################################################
class SetupProjectError(RuntimeError):
    pass


class SetupProject:
    def __init__(self):
        self._logger = logging.getLogger("SetupProject")

        ## Dictionary to store the changes to the environment.
        #  It is initialized as a copy of the current environment.
        self.environment = TemporaryEnvironment(dict(os.environ))

        self.opts = None
        self.args = None

        self.project_name = None
        self.project_version = None
        self.project_info = None

        self.context_path = None

        self.output_file = sys.stdout
        self.tmp_root = None

        if not self.check_environment():
            raise RuntimeError("Environment sanity check failed")

        self.default_externals = {
            "none": [],
            "CERN": [],  # ['CASTOR'],
        }
        self.parser = self._prepare_parser()
        self.overriding_projects = None
        self.runtime_projects = None
        self.search_path = None
        self.user_area = None

    def __getattr__(self, attr):
        """
        Shortcut to retrieve the options
        """
        if self.opts:
            return getattr(self.opts, attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )

    ## Alias to forward calls to the internal logger
    def _log(self, level, msg, *args, **kwargs):
        self._logger.log(*[level, msg] + list(args), **kwargs)

    ## Alias to forward calls to the internal logger
    def _always(self, msg, *args, **kwargs):
        self._logger.always(*[msg] + list(args), **kwargs)

    ## Alias to forward calls to the internal logger
    def _error(self, msg, *args, **kwargs):
        self._logger.error(*[msg] + list(args), **kwargs)

    ## Alias to forward calls to the internal logger
    def _warning(self, msg, *args, **kwargs):
        self._logger.warning(*[msg] + list(args), **kwargs)

    ## Alias to forward calls to the internal logger
    def _info(self, msg, *args, **kwargs):
        self._logger.info(*[msg] + list(args), **kwargs)

    ## Alias to forward calls to the internal logger
    def _verbose(self, msg, *args, **kwargs):
        self._logger.verbose(*[msg] + list(args), **kwargs)

    ## Alias to forward calls to the internal logger
    def _debug(self, msg, *args, **kwargs):
        self._logger.debug(*[msg] + list(args), **kwargs)

    ## Helper function to simplify the calls to CMT.
    def cmt(self, cmd, args=[], cwd=None):
        return cmt(cmd, args, environment=dict(self.environment), cwd=cwd)

    def check_environment(self):
        """
        Check that the current environment is usable.
        """
        self._debug("----- check_environment() -----")
        if "BINARY_TAG" in self.environment:
            if (
                "CMTCONFIG" in self.environment
                and self.environment["CMTCONFIG"] != os.environ["BINARY_TAG"]
            ):
                self._warning("both CMTCONFIG and BINARY_TAG set, using BINARY_TAG")
            self.environment["CMTCONFIG"] = os.environ["BINARY_TAG"]
        if "CMTCONFIG" not in self.environment:
            self._error("You do not have CMTCONFIG set")
            return False
        return True

    def set_build_env(self):
        self._debug("----- set_build_env() -----")
        if not self.project_info:
            # do nothing
            # FIXME: is this an error?
            return
        # make sure we have CMT in the path
        from LbEnv import which

        if not which("cmt"):
            self._debug('"cmt" not in the default environemnt, ' "I look for it...")
            import LbPlatformUtils

            from LbEnv.Bootstrap import bin_path

            cmt_cmd = which(
                "cmt",
                path=os.pathsep.join(
                    # this is a convoluted way of deducing the LbEnv standard
                    # PATH from CMAKE_PREFIX_PATH
                    bin_path(
                        [
                            os.path.dirname(d)
                            for d in self.search_path
                            if d.endswith("lhcb")
                        ],
                        host_os=LbPlatformUtils.host_os(),
                        host_flavour="{0}-{4}".format(*os.uname()),
                    )
                ),
            )
            if cmt_cmd:
                self._debug("using %s", cmt_cmd)
                self.environment["PATH"] = os.pathsep.join(
                    [self.environment["PATH"], os.path.dirname(cmt_cmd)]
                )
            else:
                self._error("I cannot find cmt")
                raise RuntimeError("Environment sanity check failed")

        if self.project_info.policy == "old":  # old style project
            self._debug("----- old style project -----")
            if "CMTPROJECTPATH" in self.environment:
                self._debug("----- unsetenv CMTPROJECTPATH -----")
                del self.environment["CMTPROJECTPATH"]
            # check if we have ExtraPackages in the override_projects
            ep_pi = None
            for pi in self.overriding_projects:
                if pi.name == "ExtraPackages":
                    ep_pi = pi
                    break
            if ep_pi:
                # if ExtraPackages is there, I prepend it to the CMTPATH
                if "CMTPATH" in self.environment:
                    self.environment["CMTPATH"] = os.pathsep.join(
                        [self.environment["CMTPATH"], ep_pi.project_dir]
                    )
                else:
                    self.environment["CMTPATH"] = ep_pi.project_dir
            # get all the variables from <Project>Env
            localEnv = dict(self.environment)
            ShellParser[self.shell](
                self.cmt(
                    "setup", "-" + self.shell, self.project_info._projectenv_cmt_dir
                ),
                localEnv,
            )
            # get the CMTPATH from <Project>Env without variable expansion
            cmtpath = os.popen("cmt show set CMTPATH").readlines()[-1].strip()
            if cmtpath.startswith("CMTPATH="):  # remove head of the line
                cmtpath = cmtpath[8:]
            cmtpath = cmtpath.strip("'")  # remove quotes
            # expand (all) the environment variables and set CMTPATH
            cmtpath_expanded = ""
            while cmtpath_expanded != cmtpath:
                if cmtpath_expanded:
                    cmtpath = cmtpath_expanded
                cmtpath_expanded = smartExpandVarsPath(cmtpath, localEnv)
            self.environment["CMTPATH"] = cmtpath
            # prepend User_release_area if defined
            if self.user_area:
                if self.environment["CMTPATH"].find(self.user_area) < 0:
                    self.environment["CMTPATH"] = os.pathsep.join(
                        [self.user_area, self.environment["CMTPATH"]]
                    )
            self._debug("----- CMTPATH set to '%s' -----", self.environment["CMTPATH"])
        else:
            self.environment["CMTPROJECTPATH"] = os.pathsep.join(self.search_path)
            self._debug(
                "----- CMTPROJECTPATH set to '%s' -----",
                self.environment["CMTPROJECTPATH"],
            )

            # unset CMTPATH if present
            if "CMTPATH" in self.environment:
                self._debug("----- unsetenv CMTPATH -----")
                del self.environment["CMTPATH"]

    def _write_script(self, data):
        close_output = False
        if self.opts.output:
            if self.append:
                self.output_file = open(self.opts.output, "a")
            else:
                self.output_file = open(self.opts.output, "w")
            self.opts.output = None  # reset the option value to avoid to reuse it
            close_output = True
        elif self.opts.mktemp:
            fd, outname = mkstemp()
            self.output_file = os.fdopen(fd, "w")
            print(outname)
            self.opts.mktemp = None  # reset the option value to avoid to reuse it
            close_output = True
        # write the data
        self.output_file.write(data)
        self.output_file.write("\n")  # @TODO: this may be avoided
        if close_output:
            self.output_file.close()

    def _prepare_parser(self):
        from optparse import IndentedHelpFormatter as BasicHelpFormatter
        from optparse import OptionParser, OptionValueError

        class MyHelpFormatter(BasicHelpFormatter):
            def format_option(self, option):
                if option.help and option.help.find("(internal)") != -1:
                    return ""
                return BasicHelpFormatter.format_option(self, option)

            def format_option_strings(self, option):
                """Return a comma-separated list of option strings & metavariables."""
                if option.takes_value():
                    metavar = option.metavar or option.dest.upper()
                    short_opts = [sopt + " " + metavar for sopt in option._short_opts]
                    long_opts = [lopt + " " + metavar for lopt in option._long_opts]
                else:
                    short_opts = option._short_opts
                    long_opts = option._long_opts

                if self.short_first:
                    opts = short_opts + long_opts
                else:
                    opts = long_opts + short_opts

                return ", ".join(opts)

        class MyOptionParser(OptionParser):
            def print_help(self, stream=None):
                if stream is None:
                    stream = sys.stderr
                return OptionParser.print_help(self, stream)

            def print_version(self, stream=None):
                if stream is None:
                    stream = sys.stderr
                return OptionParser.print_version(self, stream)

        parser = MyOptionParser(
            usage="%prog [options] <project_name> [version|--ask] [options] [externals]",
            version=__version__,
            formatter=MyHelpFormatter(),
        )

        def check_output_options_cb(option, opt_str, value, parser):
            if opt_str == "--mktemp":
                if parser.values.output:
                    raise OptionValueError(
                        "--mktemp cannot be used at the same time as --output or --append"
                    )
                parser.values.mktemp = True
            elif opt_str in ["--output", "--append"]:
                if parser.values.mktemp:
                    raise OptionValueError(
                        "--mktemp cannot be used at the same time as --output or --append"
                    )
                parser.values.output = value
                parser.values.append = opt_str == "--append"

        # internal options
        parser.add_option(
            "--shell",
            action="store",
            type="choice",
            metavar="SHELL",
            choices=["csh", "sh", "bat"],
            help="(internal) select the type of shell to use",
        )
        parser.add_option(
            "--output",
            action="callback",
            metavar="FILE",
            type="string",
            callback=check_output_options_cb,
            help="(internal) output the command to set up the environment ot the given file instead of stdout",
        )
        parser.add_option(
            "--append",
            action="callback",
            metavar="FILE",
            type="string",
            callback=check_output_options_cb,
            help="(internal) same as --output, but do not truncate",
        )
        parser.add_option(
            "--mktemp",
            action="callback",
            callback=check_output_options_cb,
            help="(internal) send the output to a temporary file and print on stdout the file name (like mktemp)",
        )

        def default_externals_cb(option, opt_str, value, parser):
            parser.values.site_externals = self.default_externals[value]

        # public options
        parser.add_option(
            "--site",
            action="callback",
            type="choice",
            metavar="SITE",
            choices=list(self.default_externals.keys()),
            callback=default_externals_cb,
            help="enable site specific defaults",
        )
        parser.add_option(
            "--ask",
            action="store_true",
            help="ask for the version of the project to use (overrides the version specified)",
        )
        parser.add_option(
            "--disable-CASTOR",
            action="store_true",
            help="remove CASTOR from the added dependencies",
        )
        parser.add_option("--tag_add", action="append", help="specify extra CMT tags")
        parser.add_option("--use", action="append", help="add a CMT use statement")
        parser.add_option(
            "--verbose",
            action="store_const",
            dest="loglevel",
            const=VERBOSE,
            help="be a bit more verbose",
        )
        parser.add_option(
            "--debug",
            action="store_const",
            dest="loglevel",
            const=DEBUG,
            help="output useful for debugging",
        )
        parser.add_option(
            "--ignore-missing",
            action="store_true",
            dest="force",
            help="obsolete, equivalent to --force",
        )
        parser.add_option(
            "--force",
            action="store_true",
            help="ignore warnings and errors from CMT, just complain",
        )
        parser.add_option(
            "--ignore-context",
            action="store_true",
            help="do not use CMTUSERCONTEXT even if it should be used",
        )
        parser.add_option(
            "--list-versions",
            action="store_true",
            help="print available versions of the specified project and exit (all other options are ignored)",
        )
        parser.add_option(
            "--external-only",
            action="store_true",
            help="sets only the environment for the externals (the project is used only to select the version of LCG)",
        )

        def dev_dir_cb(option, opt_str, value, parser):
            if value is None:
                if "LHCBDEV" in self.environment:
                    value = self.environment["LHCBDEV"]
                else:
                    raise OptionValueError("--dev used, but LHCBDEV is not defined")
            parser.values.dev_dirs.append(value)

        parser.add_option(
            "--dev",
            action="callback",
            callback=dev_dir_cb,
            help="prepend $LHCBDEV to the search path. "
            + "Note: the directories are searched in the "
            + "order specified on the command line.",
        )
        parser.add_option(
            "--dev-dir",
            action="callback",
            metavar="DEVDIR",
            type="string",
            callback=dev_dir_cb,
            help="prepend DEVDIR to the search path. "
            + "Note: the directories are searched in the "
            + "order specified on the command line.",
        )

        parser.add_option(
            "--user-area",
            action="store",
            help="Use the specified path as User_release_area instead of "
            "the value of the environment variable.",
        )

        def siteroot_option(option, opt_str, value, parser):
            from LbEnv.Bootstrap import collect_roots, search_path

            parser.values.siteroot = value
            parser.values.dev_dirs.extend(search_path(collect_roots(value)))
            self.environment["SITEROOT"] = value  # required by old LCGCMT

        parser.add_option(
            "-r",
            "--siteroot",
            type="string",
            action="callback",
            callback=siteroot_option,
            help="path to the installation root, used to add default search " "path",
        )

        def unset_option(option, opt_str, value, parser):
            if value in self.environment:
                del self.environment[value]

        parser.add_option(
            "-u",
            "--unset",
            type="string",
            action="callback",
            callback=unset_option,
            help="unset an variable before preparing the environment",
        )

        def external_version_option(_option, opt_str, value, parser):
            if len(parser.largs) < 2:
                raise OptionValueError(
                    "%s must be preceded by the name of the external" % opt_str
                )
            parser.values.ext_versions[parser.largs.pop()] = value

        parser.add_option(
            "-v",
            action="callback",
            metavar="VERSION",
            type="string",
            callback=external_version_option,
            help="must be used after the name of an external to specify a non default version for it",
        )

        parser.add_option(
            "--set-CMTPATH",
            action="store_true",
            help="Set CMTPATH to the value used internally by CMT (DANGEROUS)",
        )

        def runtime_project_option(_option, opt_str, _value, parser):
            if len(parser.rargs) < 1:
                raise OptionValueError(
                    "%s must be followed by the project name and optionally by the version"
                    % opt_str
                )
            if "/" in parser.rargs[0]:
                p_name, v = parser.rargs.pop(0).split("/")
                parser.values.runtime_projects.append((p_name, v))
                return
            p_name = parser.rargs.pop(0)
            # allow version strings of the LHCb style, LCGCMT style or glob patterns cadidates
            if len(parser.rargs) and re.match(
                r"^(v[0-9]+r[0-9]+(p[0-9]+)?)|([0-9]+[a-z]?)|(.*[\[\]\*\?].*)$",
                parser.rargs[0],
            ):
                v = parser.rargs.pop(0)
            else:
                v = None
            parser.values.runtime_projects.append((p_name, v))

        parser.add_option(
            "--runtime-project",
            action="callback",
            metavar="PROJECT [VERSION]",
            type="string",
            callback=runtime_project_option,
            nargs=0,
            help="Add a project to the runtime environment",
        )

        def overriding_project_option(_option, opt_str, _value, parser):
            if len(parser.rargs) < 1:
                raise OptionValueError(
                    "%s must be followed by the project name and optionally by the version"
                    % opt_str
                )
            if "/" in parser.rargs[0]:
                p_name, v = parser.rargs.pop(0).split("/")
                parser.values.overriding_projects.append((p_name, v))
                return
            p_name = parser.rargs.pop(0)
            if len(parser.rargs) and re.match(
                r"^(v[0-9]+r[0-9]+(p[0-9]+)?)|([0-9]+[a-z]?)|(.*[\[\]\*\?].*)$",
                parser.rargs[0],
            ):
                v = parser.rargs.pop(0)
            else:
                v = None
            parser.values.overriding_projects.append((p_name, v))

        parser.add_option(
            "--overriding-project",
            action="callback",
            metavar="PROJECT [VERSION]",
            type="string",
            callback=overriding_project_option,
            nargs=0,
            help="Add a project to override packages",
        )

        parser.add_option(
            "--no-auto-override",
            action="store_false",
            dest="auto_override",
            help="Do not automatically prepend the projects %s"
            % auto_override_projects,
        )

        parser.add_option(
            "--use-grid",
            action="store_true",
            help="Enable auto selection of LHCbGrid project",
        )

        parser.add_option(
            "-q",
            "--quiet",
            action="store_true",
            help="Removes message printout during setup",
        )

        parser.add_option(
            "--silent", action="store_true", help="Avoid any printout (even errors)"
        )

        parser.add_option(
            "--keep-CMTPROJECTPATH",
            action="store_true",
            help="Do not override the value of the environment variable CMTPROJECTPATH",
        )

        def nightly_base(option, opt_str, value, parser):
            """
            Callback for the --nightly-base and --nightly-cvmfs options.
            """
            if parser.values.nightly:
                raise OptionValueError("%s specified after --nightly" % option)

            if option.get_opt_string() == "--nightly-cvmfs":
                path = "/cvmfs/lhcbdev.cern.ch/nightlies"
            else:
                path = value
            if not os.path.isdir(path):
                raise OptionValueError('"%s" is not a directory' % path)

            global nightly_bases
            nightly_bases.append(path)

        parser.add_option(
            "--nightly-base",
            action="callback",
            type="string",
            callback=nightly_base,
            help="add the specified directory to the nightly builds "
            "search path (must be specified before --nightly)",
        )

        parser.add_option(
            "--nightly-cvmfs",
            action="callback",
            callback=nightly_base,
            help="looks for nightly builds on CVMFS "
            "(must be specified before --nightly)",
        )

        def nightly_option(_option, opt_str, _value, _parser):
            valid_value = re.compile(
                r"^(mon|tue|wed|thu|fri|sat|sun|today|yesterday|latest|"
                r"\d{4}-\d\d-\d\d|\d+)$",
                re.IGNORECASE,
            )
            day = "Today"

            parser.values.dev = True
            rargs = parser.rargs

            try:
                slot = rargs.pop(0)
            except IndexError:
                raise OptionValueError(
                    "%s must be followed by the slot of the "
                    "nightlies and optionally by the build id" % opt_str
                )

            if "/" in slot:
                slot, day = slot.split("/", 1)
                if valid_value.match(day):
                    day = day.capitalize()
            elif rargs:
                match = valid_value.match(rargs[0])
                if match:
                    day = rargs.pop(0).capitalize()
                    import logging

                    logging.warning(
                        'deprecated slot id specification: use "... '
                        '--nightly %s/%s ..." instead',
                        slot,
                        day,
                    )
            if day == "Latest":  # special case
                day = "latest"

            # Locate the requested slot in the know nightlies directories
            global nightly_bases
            nightly_bases += [
                self.environment.get(
                    "LHCBNIGHTLIES", "/cvmfs/lhcbdev.cern.ch/nightlies"
                ),
                self.environment.get(
                    "LCG_nightlies_area", "/cvmfs/sft-nightlies.cern.ch/lcg/nightlies"
                ),
            ]
            slot_dir = None
            for nightly_base in nightly_bases:
                slot_dir = os.path.join(nightly_base, slot)
                if os.path.isdir(slot_dir):
                    break  # exit from the loop as soon as the slot is found
            if not slot_dir:
                raise OptionValueError(
                    f"Cannot find slot {slot} in {nightly_bases}. Check the values of the option {opt_str}"
                )
            path = os.path.join(slot_dir, day)
            if not os.path.isdir(path):
                raise OptionValueError(
                    f"The directory {path} does not exists. Check the values of the option {opt_str}"
                )
            parser.values.dev_dirs.append(path)
            # Get the extra CMTPROJECTPATH entries needed for the nightlies
            extraCMTPROJECTPATH = None
            confSumm_file = os.path.join(path, "confSummary.py")
            config_file = os.path.join(path, "configuration.xml")
            if os.path.exists(confSumm_file):  # Try with the python digested version
                import imp

                with open(confSumm_file) as m:
                    data = imp.load_module(
                        "confSummary", m, confSumm_file, (".py", "U", 1)
                    )
                if hasattr(data, "cmtProjectPathList"):
                    # Get the list and convert it to strings
                    extraCMTPROJECTPATH = [str(d) for d in data.cmtProjectPathList if d]
                else:
                    extraCMTPROJECTPATH = []
            elif os.path.exists(config_file):  # Try with the XML configuration
                extraCMTPROJECTPATH = getNightlyCMTPROJECTPATH(config_file, slot, day)
            if extraCMTPROJECTPATH:
                parser.values.dev_dirs += extraCMTPROJECTPATH

        parser.add_option(
            "--nightly",
            action="callback",
            metavar="SLOT [DAY]",
            type="string",
            callback=nightly_option,
            nargs=0,
            help="Add the required slot of the LHCb nightly "
            + "builds to the list of DEV dirs. DAY must be "
            + "a 3 digit abbreviation of the weekday, "
            + "by default the current day. Special settings "
            + "of the CMTPROJECTPATH needed for the "
            + "nightly build slot are taken into account.",
        )

        parser.add_option(
            "--no-user-area",
            action="store_true",
            help="Ignore the user release area when looking for projects.",
        )

        parser.add_option(
            "--no-touch-logfile",
            action="store_false",
            dest="touch_logfile",
            default=True,
            help="Avoid touching the logfiles used to identify active projects.",
        )

        parser.add_option(
            "--ignore-not-ready",
            action="store_true",
            help="Ignore the presence of the NOT_READY file in a project.",
        )

        parser.add_option(
            "--profile",
            action="store_true",
            help="Print some profile informations about the execution.",
        )

        parser.add_option(
            "--ext", action="append", help="add an external lcg tool environment"
        )

        parser.add_option(
            "--use-sp",
            action="store_true",
            help="unused, needed for compatibility with lb-run",
        )

        parser.set_defaults(
            output=None,
            mktemp=False,
            append=False,
            loglevel=WARNING,
            disable_CASTOR=False,
            dev_dirs=[],
            ext_versions={},
            use=[],
            tag_add=[],
            set_CMTPATH=False,
            runtime_projects=[],
            overriding_projects=[],
            auto_override=True,
            use_grid=False,
            quiet=False,
            silent=False,
            keep_CMTPROJECTPATH=False,
            no_user_area=False,
            ext=[],
        )

        try:
            from LbEnv.ProjectEnv.options import addPlatform

            addPlatform(parser)
        except AssertionError:
            # LbEnv.ProjectEnv works only with Python >= 2.6, but we need to
            # support 2.4 too (SLC5), so we ignore the assertion
            parser.set_defaults(platform=None)

        if (
            "CMTSITE" in self.environment
            and self.environment["CMTSITE"] in self.default_externals
        ):
            parser.set_defaults(
                site_externals=self.default_externals[self.environment["CMTSITE"]]
            )
        else:
            parser.set_defaults(site_externals=self.default_externals["none"])

        return parser

    def _print_versions(self, versions):
        output = ""
        if not self.opts.quiet:
            for _p, v, _n, d in SortVersionTuples(versions):
                output += f"echo {v} in {d}\n"
        self._write_script(output)

    def _ask_version(self, versions):
        # ask for a version
        versions = SortVersions([v[1] for v in versions])
        # string representations, to handle correctly projects without version
        str_versions = [str(v) for v in versions]
        ans = None
        while not ans:
            sys.stderr.write(
                "Please enter your choice ({} q[uit] [{}]): ".format(
                    " ".join(str_versions), str_versions[-1]
                )
            )
            sys.stderr.flush()
            ans = sys.stdin.readline().strip()
            if ans == "":
                ans = versions[-1]
            elif ans.lower() in ["q", "quit"]:
                return "quit"
            elif ans not in str_versions:
                self._always("Version '%s' not valid!", ans)
                ans = None
        if ans:
            self._always("Trying version '%s'" % ans)
            # handle project without version
            if ans == "None" and None in versions:
                ans = None
        return ans

    def _touch_project_logfiles(self):
        touchline = ""
        self._debug("----- _touch_project_logfiles() -----")
        if self.shell in ["csh", "sh"]:
            # I have to touch a file to tell the release manager which version of the project I'm using
            if (
                "LHCB_USERLOGS" in self.environment
                and "USER" in self.environment
                and self.project_info.version
                and lhcb_style_version.match(self.project_info.version)
            ):  # I do not want to record non-standard or no versions
                dirname = self.environment["LHCB_USERLOGS"]
                if os.path.isdir(dirname):
                    try:
                        subdir = os.path.join(dirname, self.project_info.name.upper())
                        if not os.path.exists(subdir):
                            os.mkdir(subdir)
                        touchline = "touch {}/{}/{}_{}_{} >& /dev/null\n".format(
                            dirname,
                            self.project_info.name.upper(),
                            self.project_info.name.upper(),
                            self.project_info.version,
                            self.environment["USER"],
                        )
                    except OSError:
                        # if we cannot create the directory, we do not touch the
                        # file
                        pass
        return touchline

    def _prepare_tmp_local_project(self):
        # prepare temporary local project directory
        self.tmp_root = TempDir(prefix="SetupProject", keep_var="SPKEEPTEMPDIR")

        tmp_dir = os.path.join(
            str(self.tmp_root), "v23"
        )  # add a fake version directory to please CMT
        os.mkdir(tmp_dir)

        self._debug("Using temporary directory '%s'" % tmp_dir)

        # prepare temporary requirements file
        req = open(os.path.join(tmp_dir, "requirements"), "w")

        # prepare use statementes for user-requested packages
        use_rexp = re.compile("^(?:([^/]*)/)?([^/ .]*)[ .]*([^ ]+)? *$")
        for u in self.opts.use:
            m = use_rexp.match(u)
            if (
                m
            ):  # handle the case of "use" string in the format "[Hat/]Package [version]"
                hat, pack, ver = m.groups()
                if ver is None:
                    ver = "v*"
                if hat is None:
                    from LbEnv import getPackageNames

                    suffix = "/" + pack
                    hat = ""
                    for p in getPackageNames():
                        if p.endswith(suffix):
                            hat = p.split("/", 1)[0]
                            break
                    if hat:
                        self._warning(
                            "wrong data package name used (%s), the"
                            " correct name is %s/%s",
                            pack,
                            hat,
                            pack,
                        )
                req.write(f"use {pack} {ver} {hat}\n")
            else:  # for any other case, just use the provided string
                req.write("use %s\n" % u)

        if not self.external_only and self.project_info.sys:
            # if we were not asked only for external and the Sys package exists,
            # we use it
            req.write("use %s *\n" % self.project_info.sys)

        # add user specified tags
        for t in self.tag_add:
            req.write("apply_tag %s\n" % t)

        # prepare use statementes for externals (with default version)
        # version override have to happen later
        exts = sum([e.split(",") for e in self.opts.ext], [])
        for e in (
            exts + self.args + list(self.ext_versions.keys()) + self.site_externals
        ):
            # @todo:I do not like too much this few lines
            if e != "CASTOR" or not self.disable_CASTOR:  # skip CASTOR
                req.write("use %s v* LCG_Interfaces\n" % e)

        # set config versions
        # (the override of the versions have to be done after the package:
        #  some packages may change the default version, but the command line
        #  must win)
        for e in self.ext_versions:
            req.write(f'macro {e}_config_version "{self.ext_versions[e]}"\n')

        # use external location relative to the LCG Interfaces for the local installation

        req.write(
            'macro LCG_home "$(LCG_home)" LOCAL&Unix "$(LCG_SETTINGSROOT)/../../../.." LOCAL&WIN32 "$(LCG_SETTINGSROOT)\\..\\..\\..\\.."\n'
        )

        req.flush()
        del req

        # prepare temporary project file
        os.mkdir(os.path.join(tmp_dir, "cmt"))
        prj = open(os.path.join(tmp_dir, "cmt", "project.cmt"), "w")
        # add usage for the requested project(s)
        for p in self.overriding_projects + [self.project_info] + self.runtime_projects:
            prj.write("use %s\n" % p.realName.replace(os.sep, " "))
        prj.flush()
        del prj

        if self._logger.level <= 2:
            self._verbose("########## requirements ##########")
            self._verbose(open(os.path.join(tmp_dir, "requirements")).read())
            self._verbose("########## project.cmt ##########")
            self._verbose(open(os.path.join(tmp_dir, "cmt", "project.cmt")).read())

        return tmp_dir

    def _gen_setup(self, root_dir):
        """
        Generate the setup script.
        Returns (script,error) where script is a string and error a list.
        If script is empty, you should find the reason in error.
        """
        self._debug("----- gen_setup() -----")

        if (
            "CMTPATH" not in self.environment
            and "CMTPROJECTPATH" not in self.environment
        ):
            # it does not make sense to go on if the CMTPATH is not set
            raise SetupProjectError("neither CMTPATH nor CMTPROJECTPATH are set")

        if self.context_path:
            self.environment["CMTUSERCONTEXT"] = self.context_path

        # check CMT
        out = self.cmt("version", cwd=root_dir)
        if "command not found" in out:
            raise SetupProjectError("cmt is not in your PATH")
        else:
            self._debug("----- using CMT %s -----" % out.strip().split("\n")[-1])

        # check if the project works
        out = self.cmt("check", "configuration", cwd=root_dir)
        if out and not self.force:  # non empty output means error
            # ignore incompatibility notices
            inc_notice = re.compile(
                r"^("
                r"# Required version.*incompatible with|"
                r"#CMT---> Warning: apply_tag with empty name"
                r").*"
            )
            if [l for l in out.splitlines() if not inc_notice.match(l)]:
                raise SetupProjectError(out)

        # Remove entries containing /gcc/ from the *PATH variables
        # https://its.cern.ch/jira/browse/LBCORE-899
        for v in (
            "PATH",
            "PYTHONPATH",
            "LD_LIBRARY_PATH",
            "ROOT_INCLUDE_PATH",
            "HPATH",
            "MANPATH",
        ):
            if v in self.environment:
                self.environment[v] = os.pathsep.join(
                    x for x in self.environment[v].split(os.pathsep) if "/gcc/" not in x
                )

        script = self.cmt("setup", "-" + self.shell, cwd=root_dir)

        # parse the output
        new_env = TemporaryEnvironment(self.environment)
        # this sets the new environment end return the line it cannot understand
        script = ShellParser[self.shell](script, new_env)

        # remove the temporary directory from the paths
        root_dir_local = root_dir
        if self.opts.shell == "bat":
            root_dir_local = root_dir_local.replace("/", "\\")
        for v in ["PATH", "PYTHONPATH", "LD_LIBRARY_PATH", "HPATH", "MANPATH"]:
            if v in new_env:
                # Keep the entries that do not refer to the local temporary directory,
                # do exist and are not empty
                new_env[v] = os.pathsep.join(
                    [d for d in new_env[v].split(os.pathsep) if root_dir_local not in d]
                )
                if "LB_NO_STRIP_PATH" not in os.environ:
                    self._debug("Cleaning variable %s", v)
                    self._debug("  before: %s", new_env[v])
                    new_env[v] = StripPath(new_env[v])
                    self._debug("   after: %s", new_env[v])
        # FIXME: I should look for all the variables pointing to the temporary directory

        # remove the variables that have the temporary directory in the name
        tmp_base_name = os.path.basename(str(self.tmp_root))
        for k in [
            k
            for k in list(new_env.keys())  # I need a copy of the keys
            if tmp_base_name in k
        ]:
            del new_env[k]

        # remove CMTPATH if not explicitely requested
        if not self.set_CMTPATH and "CMTPATH" in new_env:
            del new_env["CMTPATH"]

        lines = []
        errors = []
        if self.context_path:
            if not self.opts.quiet:
                lines.append("echo Using CMTUSERCONTEXT = '%s'" % self.context_path)
            # unset CMTUSERCONTEXT in case of future calls
            del new_env["CMTUSERCONTEXT"]

        for l in script.splitlines():
            if l.startswith("#CMT"):
                errors.append(l)
            else:
                lines.append(l)
        if lines:
            lines.append("")

        if errors:
            self._debug("----- gen_setup(): errors != [] -----")
        # Do not forget changes in the environment
        new_env.commit()
        return ("\n".join(lines), errors)

    def _prepend_prompt(self):
        new_prompt = ""
        if (
            os.environ["SP_PROMPT"] == "without_version"
            or not self.project_info.version
        ):
            if self.shell == "sh":
                new_prompt = 'if [[ -n "$PS1" ]]; then\n'
                # assume that ZSH_VERSION is always set in zsh and never in bash
                new_prompt += "  if [[ -n $ZSH_VERSION ]]; then\n"
                # zsh encodes colour codes with $'\e[...m'
                # zsh encapulates characters which don't move the coursor in %{ %}
                #          (this ensures sane behaviour with reverse search ^R)
                # the double backslashes are to escape a backslash in the pyhton string
                # %% is escaping a single % to avoid collision with formating in `".." % variable`
                new_prompt += (
                    "    PS1=$'%%{\\e[1;34m%%}[%s]%%{\\e[m%%}'$PS1\n"
                    % self.project_info.name
                )
                new_prompt += "  else\n"
                # bash encodes colour codes with $'\e[...m' (though it works without the dollar sign)
                # bash encapulates characters which don't move the coursor in \[ \]
                #          (this ensures sane behaviour with reverse search ^R)
                # the double backslashes are to escape a backslash in the pyhton string
                new_prompt += (
                    "    PS1=$'\\[\\e[1;34m\\][%s]\\[\\e[m\\]'$PS1\n"
                    % self.project_info.name
                )
                new_prompt += "  fi\n"
                new_prompt += "fi\n"
            elif self.shell == "csh":
                new_prompt = "if ( $?prompt ) then\n"
                new_prompt += (
                    '  set prompt="%%B%%{\\033[34m%%}[%s]%%{\\033[0m%%}%%b$prompt"\n'
                    % self.project_info.name
                )
                new_prompt += "endif\n"
        else:
            if self.shell == "sh":
                new_prompt = 'if [[ -n "$PS1" ]]; then\n'
                # assume that ZSH_VERSION is always set in zsh and never in bash
                new_prompt += "  if [[ -n $ZSH_VERSION ]]; then\n"
                new_prompt += (
                    "    PS1=$'%{{\\e[1;34m%}}[{} {}]%{{\\e[m%}}'$PS1 \n".format(
                        self.project_info.name,
                        self.project_info.version,
                    )
                )
                # new_prompt += "    echo 'setting prompt for zsh'\n"
                new_prompt += "  else\n"
                new_prompt += (
                    "    PS1=$'\\[\\e[1;34m\\][{} {}]\\[\\e[m\\]'$PS1 \n".format(
                        self.project_info.name,
                        self.project_info.version,
                    )
                )
                # new_prompt += "    echo 'setting prompt for bash'\n"
                new_prompt += "  fi\n"
                new_prompt += "fi\n"
            elif self.shell == "csh":
                new_prompt = "if ( $?prompt ) then\n"
                new_prompt += f'  set prompt="%B%{{\\033[34m%}}[{self.project_info.name} {self.project_info.version}]%{{\\033[0m%}}%b$prompt"\n'
                new_prompt += "endif\n"

        return new_prompt

    def parse_args(self, args=sys.argv[1:]):
        self.opts, self.args = self.parser.parse_args(args=args)
        # It's not possible to use callbacks with default values, so we need to
        # handle the default value for the --siteroot option here.
        if self.opts.siteroot is None:
            self.parser.get_option("--siteroot").callback(
                self.parser.get_option("--siteroot"),
                "--siteroot",
                computeDefaultSiteRoot(),
                self.parser,
            )
        # override parse_args to ignore --use-grid
        self.opts.use_grid = False

    def prepare(self, args=None):
        """
        Collect all the informations needed to set up the environment, like
        project name and version, paths, etc.
        """
        # If the option --silent is specified, --quiet is implied
        if self.opts.silent:
            self.opts.quiet = True

        # set level of log messages
        if not self.opts.quiet:
            self._logger.level = self.loglevel
        elif self.opts.silent:
            self._logger.level = ALWAYS + 10  # do not even print ALWAYS messages
        else:  # quiet but not silent
            self._logger.level = ERROR
        log.level = self.loglevel

        self._debug("----- main() -----")

        self._verbose("SetupProject %s" % self.parser.get_version())

        # ------------- get project name
        if len(self.args) < 1:
            self._error("You have to specify a project")
            return 1

        required_root_version = None
        if self.args[0].lower() == "root":
            if self.list_versions:
                self._error("cannot list the versions of ROOT")
                return 1
            # ROOT is not a CMT project, instead of
            #   SetupProject ROOT 5.34.10
            # we should do
            #   SetupProject LCGCMT 66 ROOT
            self.project_name = "LCGCMT"
            # let's see if the user actually passed us a version for ROOT
            if len(self.args) > 1 and re.match(r"\d+\.\d+\.\d+", self.args[1]):
                # yes, record it for later (when we have the search path ready)
                required_root_version = self.args[1]
                # fix the arguments
                self.args[0:2] = ["ROOT"]
            else:
                self._debug("Using the latest version of LCGCMT to provide ROOT")
                # if the version is not given, we do not need to do anything
                # special on the arguments, just ensure that the case is correct
                self.args[0] = "ROOT"
        else:
            if "/" in self.args[0]:
                self.project_name, self.project_version = self.args.pop(0).split("/")
            else:
                self.project_name = self.args.pop(0)
            self.project_name = fixProjectCase(self.project_name)
            if self.project_name == "LCG":
                self.project_name = (
                    "LCGCMT"  # special case for compatibility with lb-run
                )

        # ------------- sanity check: the shell type has to be specified
        if not self.shell:
            self._error("Internal error: shell type not specified")
            return 1

        # ------------- set user area
        if self.opts.no_user_area:
            self.user_area = None
        elif self.opts.user_area:
            self.user_area = self.opts.user_area
        else:
            self.user_area = self.environment.get("User_release_area", None)

        # ------------- prepare search_path
        self.search_path = []
        # user area
        if self.user_area:
            self.search_path.append(self.user_area)
        # dev dirs
        if self.dev_dirs:
            self.search_path += self.dev_dirs
        # default locations
        projpathvars = ["CMAKE_PREFIX_PATH", "LHCBPROJECTPATH"]
        if self.keep_CMTPROJECTPATH:
            projpathvars.insert(0, "CMTPROJECTPATH")
        for v in projpathvars:
            if v in self.environment:
                self.search_path += self.environment[v].split(os.pathsep)

        # remove duplicates
        self.search_path = uniq(self.search_path)
        # remove entries that do not exist (bug #75480)
        self.search_path = [p for p in self.search_path if os.path.exists(p)]

        # ------------- discover all project versions
        # debug printout: print project, search path and, optionally, user area
        self._debug(
            f"Look for all versions of '{self.project_name}' in {self.search_path}"
            + ((self.user_area and (" with user area in '%s'" % self.user_area)) or "")
        )
        versions = FindProjectVersions(
            self.project_name,
            self.search_path,
            self.user_area,
            self.opts.ignore_not_ready,
            self.project_version,
        )

        if not versions:
            self._error("Cannot find project '%s'" % self.project_name)
            return 1

        if self.project_version:
            for v in versions:
                if VersionMatch(v[1], self.project_version):
                    self.project_version = v[1]
                    break
            else:
                self._error(
                    f"Cannot find version '{self.project_version}' of {self.project_name}. Try with --list."
                )
                return 1

        # Now I can print the version if I was asked to do it
        if self.list_versions:
            self._print_versions(versions)
            return 0

        if required_root_version:
            # now we can look for the version of LCGCMT that matches the required
            # version of ROOT
            self._debug(
                "Looking for the LCGCMT version providing ROOT %s"
                % required_root_version
            )
            lcg_version = FindLCGForROOT(
                required_root_version, SortVersionTuples(versions, reverse=True)
            )
            if lcg_version:
                # we need the version in the args because it's processed later
                self.args.insert(0, lcg_version)
            else:
                self._error("Cannot find ROOT %s" % required_root_version)
                return 1

        # ------------- project version
        if self.args and not self.project_version:
            # check if the next argument can be interpreted as a version for the requested project
            found = False
            pattern = self.args[0]
            if pattern.lower() == "latest":
                self.args.pop(0)
                self.project_version = None
            else:
                if pattern.upper() == "HEAD":
                    pattern = self.args[0] = "HEAD"
                for v in versions:
                    if VersionMatch(v[1], pattern):
                        found = True
                        break
                if found:
                    self.project_version = self.args.pop(0)
                elif re.match("v[0-9]*r[0-9p]*|HEAD", self.args[0]):
                    self._error(
                        f"Cannot find version '{self.args[0]}' of {self.project_name}. Try with --list-versions."
                    )
                    return 1

        # ------------- Prompt for a version to use
        if self.ask:
            v = self._ask_version(versions)
            if v == "quit":
                self._always("Quit")
                return 1
            self.project_version = v

        # ------------- Initialize the ProjectInfo objects
        # Main project
        self.project_info = makeProjectInfo(
            version=self.project_version,
            versions=versions,
            env=self.environment,
            ignore_not_ready=self.opts.ignore_not_ready,
        )
        if not self.project_info:
            # we should never get here
            self._error(
                f"PANIC: Cannot find version '{self.project_version}' of {self.project_name} after initial check"
            )
            return 1

        # runtime projects
        self.runtime_projects = []
        for p, v in self.opts.runtime_projects:
            vv = FindProjectVersions(
                p, self.search_path, self.user_area, self.opts.ignore_not_ready
            )
            if not vv:
                self._error("Cannot find project '%s'" % p)
                return 1
            pi = makeProjectInfo(
                version=v,
                versions=vv,
                env=self.environment,
                ignore_not_ready=self.opts.ignore_not_ready,
            )
            if not pi:
                self._error(
                    f"Cannot find version '{v}' of {p}. Try with --list-versions."
                )
                return 1
            self.runtime_projects.append(pi)

        # overriding projects
        self.overriding_projects = []
        for p, v in self.opts.overriding_projects:
            vv = FindProjectVersions(
                p, self.search_path, self.user_area, self.opts.ignore_not_ready
            )
            if not vv:
                self._error("Cannot find project '%s'" % p)
                return 1
            pi = makeProjectInfo(
                version=v,
                versions=vv,
                env=self.environment,
                ignore_not_ready=self.opts.ignore_not_ready,
            )
            if not pi:
                self._error(
                    f"Cannot find version '{v}' of {p}. Try with --list-versions."
                )
                return 1
            self.overriding_projects.append(pi)

        # use LHCbGrid
        if self.opts.use_grid:
            auto_override_projects.append(("LHCbGrid", ["LHCbGridSys"]))

        # auto-override projects
        if self.auto_override:
            for p, pkgs in auto_override_projects:
                vv = FindProjectVersions(
                    p, self.search_path, self.user_area, self.opts.ignore_not_ready
                )
                if vv:
                    self.overriding_projects.insert(
                        0,
                        makeProjectInfo(
                            versions=vv,
                            env=self.environment,
                            ignore_not_ready=self.opts.ignore_not_ready,
                        ),
                    )
                    self.opts.use += pkgs

        platform = self.opts.platform or os.environ["CMTCONFIG"]
        for p in self.overriding_projects + [self.project_info] + self.runtime_projects:
            if not p.supportsPlatform(platform, self.user_area):
                self._error(
                    f"Project {p.name} {p.version} is not available for platform {platform}"
                )
                return 1
            self._verbose(f"Project {p.name} {p.version} uses {p.policy} policy")

        if self.opts.platform:
            self.environment["CMTCONFIG"] = self.opts.platform

        return 0

    def main(self, args=None):
        """
        Main entry point to SetupProject script.

        Calls parse_args, then the 'run' method, profiling it if requested.
        """
        if not logging.getLogger().handlers:
            # initialize the logger (print on standard error)
            log_handler = logging.StreamHandler(sys.stderr)
            log_handler.setFormatter(logging.Formatter("%(message)s"))
            logging.getLogger().addHandler(log_handler)

        if args is None:
            args = sys.argv[1:]
        # Process commmand line options
        self.parse_args(args=args)

        # Required for some old LCGCMT
        if "SITEROOT" not in self.environment and "MYSITEROOT" in self.environment:
            self.environment["SITEROOT"] = self.environment["MYSITEROOT"]
        # this means not-AFS (the only option at the moment)
        self.environment["CMTSITE"] = "LOCAL"

        # Prepare the profiler instance
        profiler = None
        if self.opts.profile:
            try:
                import cProfile

                profiler = cProfile.Profile()
            except ImportError:
                try:
                    import profile

                    profiler = profile.Profile()
                except ImportError:
                    self._warning(
                        "Cannot import cProfile or profile: ignoring --profile"
                    )
        # if we managed to get the profiler instance, collect profiling stats,
        # and print them on stderr
        if profiler:
            rc = profiler.runcall(self.run)
            # adapted from Python standard profiler.py
            import pstats

            stats = pstats.Stats(profiler, stream=sys.stderr)
            stats.strip_dirs().sort_stats("cumulative").print_stats(10)
        else:
            rc = self.run()

        return rc

    def run(self, args=None):
        # Initialization from arguments
        rc = self.prepare(args)
        if rc or self.list_versions:  # No need to go on if --list-versions or error
            return rc

        # Prepare the build-time environment
        self.set_build_env()
        # We usually unset CMTPATH, but we need to remember it for old style
        # projects
        CMTPATH = self.environment.get("CMTPATH", "")

        script = ""  # things we need to append to the setup script (like aliases)
        messages = []  # lines to print (feedback)

        if not self.opts.quiet:
            self._always("Configuring %s" % self.project_info)
        tmp_dir = self._prepare_tmp_local_project()
        try:
            (script, err) = self._gen_setup(tmp_dir)
            if err and self.force:
                self._always("WARNING - ignoring warnings and errors from CMT:")
                self._always("\n".join(err))
        except SetupProjectError as x:
            self._error(str(x))
            if "package CASTOR" in str(x):
                self._error("Tip: try adding --disable-CASTOR to the options")
            self._error("Could not produce the environment, check the arguments")
            return 1
        # Feedback
        if "CMTPROJECTPATH" in self.environment:
            messages.append(
                "Using CMTPROJECTPATH = '{CMTPROJECTPATH}'".format(**self.environment)
            )
            if self.keep_CMTPROJECTPATH:
                messages.append("(as defined by the user)")
        else:
            messages.append("Using CMTPATH = '%s'" % CMTPATH)
        tmps = self.project_info.name
        if self.project_info.version:
            tmps += " %s" % self.project_info.version
        messages.append("Environment for %s ready." % tmps)

        # collect project infos descriptions
        lines = []
        for pi in self.overriding_projects:
            lines.append(str(pi))
        lines.append(str(self.project_info))
        for pi in self.runtime_projects:
            lines.append(str(pi))
        # add decorators
        if len(lines) > 1:
            lines = (
                ["(%s," % lines[0]]
                + [" %s," % l for l in lines[1:-1]]
                + [" %s)" % lines[-1]]
            )
        else:
            lines[0] = "(%s)" % lines[0]
        messages += lines

        # FIXME: Hack to hide the fact that old projects were not setting the PATH for the executable
        if self.project_info.policy == "old":
            varname = self.project_info.name.upper() + "ROOT"
            if varname in self.environment:
                exedir = os.path.join(self.environment[varname], self.opts.platform)
                if os.path.isdir(exedir):
                    # it make sense to add it only if it exists
                    messages.append("Appending %s to the path." % exedir)
                    self.environment["PATH"] = os.pathsep.join(
                        [self.environment["PATH"], exedir]
                    )

        # normalize relative path introduced for the location the externals relatively
        # to the LCGCMT project
        path2normalize = ["PYTHONPATH", "PATH", "LD_LIBRARY_PATH", "ROOTSYS"]
        for p in path2normalize:
            if p in self.environment:
                pthlist = self.environment[p].split(os.pathsep)
                newlist = []
                for l in pthlist:
                    newpath = os.path.normpath(l)
                    if os.path.exists(newpath) and l.find("..") != -1:
                        newlist.append(newpath)
                    else:
                        newlist.append(l)
                self.environment[p] = os.pathsep.join(newlist)

        output_script = self.environment.gen_script(self.shell)
        output_script += script
        if not self.opts.quiet:
            for m in messages:
                output_script += 'echo "%s"\n' % m
        if self.opts.touch_logfile:
            # I have to touch a file to tell the release manager which version of the project I'm using
            output_script += self._touch_project_logfiles()

        if "SP_PROMPT" in os.environ:
            output_script += self._prepend_prompt()

        self._verbose("########## done ##########")

        # as very last step, generate the output script to be sourced by the wrapper
        self._write_script(output_script)

        ### completed successfully
        return 0


class Comp(SetupProject):
    def _prepare_parser(self):
        # add some extra (ignored) options
        parser = super()._prepare_parser()
        parser.add_option(
            "--sh",
            action="store_const",
            const="sh",
            dest="outshell",
            help="Print the environment as shell commands " "for 'sh'-derived shells.",
        )
        parser.add_option(
            "--csh",
            action="store_const",
            const="csh",
            dest="outshell",
            help="Print the environment as shell commands " "for 'csh'-derived shells.",
        )
        parser.add_option(
            "--py",
            action="store_const",
            const="py",
            dest="outshell",
            help="Print the environment as Python " "dictionary.",
        )
        parser.add_option(
            "-A",
            "--all",
            action="store_true",
            help="Print all variables, instead of just the changes, with --sh, --csh and --py.",
        )
        parser.add_option("--container", help="ignored")
        parser.add_option("--allow-containers", action="store_true", help="ignored")
        parser.add_option("--disallow-containers", action="store_true", help="ignored")
        parser.add_option("--prefer-container", action="store_true", help="ignored")
        parser.add_option("--bind", action="append", help="ignored")
        return parser

    def _write_script(self, data):
        # disable writing of the shell script
        pass


def expandAllVars(env, iterations=10):
    """
    Expand all variables in the environment dict.

    >>> env = {'MYDATA': '${DATA}/subdir', 'DATA': '/main/path'}
    >>> expandAllVars(env)['MYDATA']
    '/main/path/subdir'
    """
    from string import Template

    def expandVars(value, env):
        """expand variables in a string using env, ignoring format errors"""
        try:
            return Template(value).safe_substitute(env)
        except ValueError:
            return value

    new_env = {key: expandVars(value, env) for key, value in env.items()}
    if new_env != env:  # we did some change
        if iterations > 0:
            # let's see if there is something more to change
            return expandAllVars(new_env, iterations - 1)
        else:
            raise ValueError("failed to resolve all the variables (loop?)")
    return new_env


def getOldEnvironment(args):
    comp = Comp()
    rc = comp.main(["--shell", "sh"] + args)
    if rc:
        sys.exit(rc)
    return expandAllVars(dict(comp.environment.items()))



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################


__author__ = "Marco Clemencic <marco.clemencic@cern.ch>"

import os
import shutil
import sys
from tempfile import mkdtemp
from unittest.mock import patch

import pytest


@pytest.fixture(scope="session", autouse=True)
def setup():
    from os.path import dirname, join

    os.environ["CMAKE_PREFIX_PATH"] = join(dirname(__file__), "data")
    yield


latest_gaudi = "v27r1"


def test_platform_sorting():
    from random import shuffle

    from LbEnv.ProjectEnv.lookup import platform_sort_key

    expected_order = [
        "x86_64-slc6-gcc48-do0",
        "x86_64-slc6-gcc49-do0",
        "x86_64-slc6-gcc62-do0",
        "x86_64-centos7-gcc62-do0",
        "x86_64-centos7-gcc7-do0",
        "i686-slc5-gcc43-dbg",
        "x86_64-slc5-gcc43-dbg",
        "x86_64-slc5-gcc46-dbg",
        "x86_64-slc6-gcc46-dbg",
        "x86_64-slc6-gcc48-dbg",
        "x86_64-slc6-gcc49-dbg",
        "x86_64-centos7-gcc49-dbg",
        "x86_64-slc6-gcc62-dbg",
        "x86_64-centos7-gcc62-dbg",
        "x86_64-centos7-gcc7-dbg",
        "x86_64+avx2+fma-centos7-gcc62-dbg",
        "slc4_ia32_gcc34",
        "i686-slc5-gcc43-opt",
        "x86_64-centos7-clang8-opt",
        "slc4_amd64_gcc34",
        "x86_64-slc5-gcc43-opt",
        "x86_64-slc5-gcc46-opt",
        "x86_64-slc6-gcc46-opt",
        "x86_64-slc6-gcc48-opt",
        "x86_64-slc6-gcc49-opt",
        "x86_64-centos7-gcc49-opt",
        "x86_64-slc6-gcc62-opt",
        "x86_64-centos7-gcc62-opt",
        "x86_64-centos7-gcc7-opt",
        "x86_64-centos7-gcc8-opt",
        "x86_64-centos7-gcc10+py3-opt",
        "x86_64-centos7-gcc10-opt",
        "x86_64+avx2+fma-centos7-gcc62-opt",
    ]

    for _ in range(100):
        shuffled = list(expected_order)
        shuffle(shuffled)
        sorted_platforms = sorted(shuffled, key=platform_sort_key)
        assert (
            sorted_platforms == expected_order
        ), f"bad sorting: {shuffled} -> {sorted_platforms}"


def prepare_tree(base, tree):
    from os import makedirs
    from os.path import dirname, exists, join

    if hasattr(tree, "items"):
        tree = tree.items()
    for k, v in tree:
        k = join(base, k)
        if v:
            d = dirname(k)
            if not exists(d):
                makedirs(d)
            f = open(k, "w")
            f.write(v)
            f.close()
        else:
            if not exists(k):
                makedirs(k)


class TempDir:
    def __init__(self, tree=None):
        self.tmpdir = None
        self.tree = tree

    def __enter__(self):
        self.tmpdir = mkdtemp()
        if self.tree:
            prepare_tree(self.tmpdir, self.tree)
        return self.tmpdir

    def __exit__(self, type_, value, traceback):
        if self.tmpdir:
            shutil.rmtree(self.tmpdir)


def test_import():
    import LbEnv.ProjectEnv

    assert LbEnv.ProjectEnv.path


def test_version():
    from LbEnv.ProjectEnv.version import expandVersionAlias, isValidVersion

    assert expandVersionAlias("Gaudi", "latest", "x86_64-slc6-gcc46-opt") == "v25r0"

    assert isValidVersion("Gaudi", "latest")
    assert isValidVersion("Gaudi", "HEAD")
    assert isValidVersion("Gaudi", latest_gaudi)
    assert isValidVersion("LHCb", "v32r5p1")
    assert isValidVersion("LHCb", "v32r4g1")
    assert isValidVersion("LCGCMT", "64d")
    assert not isValidVersion("Gaudi", "a random string")


# FIXME this is an integration test...
def _test_compatibility():
    from LbEnv.ProjectEnv.compatibility import getOldEnvironment

    env_bk = dict(os.environ)

    try:
        os.environ["CMTCONFIG"] = "x86_64-slc6-gcc49-opt"
        if "BINARY_TAG" in os.environ:
            del os.environ["BINARY_TAG"]
        env = getOldEnvironment(["Gaudi", latest_gaudi])
        assert "GAUDISYSROOT" in env
        assert latest_gaudi in env["GAUDISYSROOT"]
    finally:
        os.environ.clear()
        os.environ.update(env_bk)

    try:
        os.environ["BINARY_TAG"] = "x86_64-slc6-gcc49-opt"
        os.environ["CMTCONFIG"] = "dummy"
        env = getOldEnvironment(["Gaudi", latest_gaudi])
        assert "GAUDISYSROOT" in env
        assert latest_gaudi in env["GAUDISYSROOT"]
    finally:
        os.environ.clear()
        os.environ.update(env_bk)

    try:
        env = getOldEnvironment(["NotAProject", "v1r0"])
        assert False, "exception expected"
    except SystemExit as x:
        assert x.code == 1
    finally:
        os.environ.clear()
        os.environ.update(env_bk)


def parse_args(func, args):
    from optparse import OptionParser

    return func(OptionParser(prog="dummy_program")).parse_args(args)


def test_options_addOutputLevel():
    import logging

    from LbEnv.ProjectEnv.options import addOutputLevel

    opts, _ = parse_args(addOutputLevel, [])
    assert opts.log_level == logging.WARNING

    opts, _ = parse_args(addOutputLevel, ["--debug"])
    assert opts.log_level == logging.DEBUG

    opts, _ = parse_args(addOutputLevel, ["--verbose"])
    assert opts.log_level == logging.INFO

    opts, _ = parse_args(addOutputLevel, ["--quiet"])
    assert opts.log_level == logging.WARNING


def test_options_addPlatform():
    from LbEnv.ProjectEnv.options import addPlatform, checkPlatform

    def parse_args(args):
        from optparse import OptionParser

        parser = addPlatform(OptionParser(prog="dummy_program"))
        opts, args = parser.parse_args(args)
        opts.platform = checkPlatform(parser, opts.platform)
        return opts, args

    # set platform, short optino
    opts, _ = parse_args(["-c", "platform1"])
    assert opts.platform == "platform1"

    # set platform, long optino
    opts, _ = parse_args(["--platform", "platform2"])
    assert opts.platform == "platform2"

    # take CMTCONFIG
    if "CMTCONFIG" not in os.environ:
        os.environ["CMTCONFIG"] = "default"
    if "BINARY_TAG" in os.environ:
        del os.environ["BINARY_TAG"]
    opts, _ = parse_args([])
    assert opts.platform == os.environ["CMTCONFIG"]

    # take BINARY_TAG
    if "BINARY_TAG" not in os.environ:
        os.environ["BINARY_TAG"] = "default"
    if "CMTCONFIG" in os.environ:
        del os.environ["CMTCONFIG"]
    opts, _ = parse_args([])
    assert opts.platform == os.environ["BINARY_TAG"]

    # allow equal BINARY_TAG and CMTCONFIG
    os.environ["BINARY_TAG"] = os.environ["CMTCONFIG"] = "default"
    opts, _ = parse_args([])
    assert opts.platform == os.environ["BINARY_TAG"]

    # fail for inconsistent
    os.environ["BINARY_TAG"] = "another"
    try:
        opts, _ = parse_args([])
        assert False, "failure expected"
    except SystemExit:
        pass
    del os.environ["CMTCONFIG"]
    del os.environ["BINARY_TAG"]

    # check default (not set)
    opts, _ = parse_args([])
    assert not opts.platform


def test_options_addSearchPath():
    from LbEnv.ProjectEnv.lookup import InvalidNightlySlotError
    from LbEnv.ProjectEnv.options import addSearchPath

    if "LHCBDEV" not in os.environ:
        os.environ["LHCBDEV"] = "/afs/cern.ch/lhcb/software/DEV"
    opts, _ = parse_args(addSearchPath, ["--dev"])
    assert os.environ["LHCBDEV"] in map(str, opts.dev_dirs)

    del os.environ["LHCBDEV"]
    try:
        opts, _ = parse_args(addSearchPath, ["--dev"])
        assert False, "exception expected"
    except SystemExit:
        pass

    opts, _ = parse_args(addSearchPath, ["--dev-dir", "/some/path"])
    assert "/some/path" in map(str, opts.dev_dirs)

    if "User_release_area" in os.environ:
        del os.environ["User_release_area"]
    opts, _ = parse_args(addSearchPath, [])
    assert opts.user_area is None

    os.environ["User_release_area"] = "/home/myself/cmtuser"
    opts, _ = parse_args(addSearchPath, [])
    assert opts.user_area == os.environ["User_release_area"]
    assert not opts.no_user_area

    opts, _ = parse_args(addSearchPath, ["--user-area", "/tmp/myself"])
    assert opts.user_area == "/tmp/myself"

    opts, _ = parse_args(addSearchPath, ["--no-user-area"])
    assert opts.no_user_area

    try:
        opts, _ = parse_args(addSearchPath, ["--nightly"])
        assert False, "exception expected"
    except SystemExit:
        pass

    import datetime

    days = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
    weekday = datetime.date.today().weekday()
    today = "Today"  # days[weekday]
    yesterday = days[(weekday - 1) % 7]

    with TempDir(
        {"lhcb-nightly-slot/%s" % today: None, "lhcb-nightly-slot/%s" % yesterday: None}
    ) as tmp:
        os.environ["LHCBNIGHTLIES"] = tmp
        opts, _ = parse_args(addSearchPath, ["--nightly", "lhcb-nightly-slot"])
        assert os.path.join(tmp, "lhcb-nightly-slot", today) in map(str, opts.dev_dirs)
        assert opts.nightly == ("lhcb-nightly-slot", today, tmp), opts.nightly
        opts, _ = parse_args(
            addSearchPath, ["--nightly", "lhcb-nightly-slot", yesterday]
        )
        assert os.path.join(tmp, "lhcb-nightly-slot", yesterday) in map(
            str, opts.dev_dirs
        )
        assert opts.nightly == ("lhcb-nightly-slot", yesterday, tmp), opts.nightly

        opts, _ = parse_args(addSearchPath, ["--nightly", "dummy"])
        assert isinstance(opts.nightly, InvalidNightlySlotError)

        opts, _ = parse_args(
            addSearchPath, ["--nightly", "lhcb-nightly-slot", days[(weekday + 1) % 7]]
        )
        assert isinstance(opts.nightly, InvalidNightlySlotError)

    with TempDir(
        {
            "lhcb-nightly-slot/%s/confSummary.py"
            % today: """cmtProjectPathList = ['/extra/dir']\n"""
        }
    ) as tmp:
        os.environ["LHCBNIGHTLIES"] = tmp
        opts, _ = parse_args(addSearchPath, ["--nightly", "lhcb-nightly-slot"])
        assert "/extra/dir" in map(str, opts.dev_dirs)

    with TempDir(
        {
            "lhcb-nightly-slot/%s/configuration.xml"
            % today: """<configuration><slot name="lhcb-nightly-slot">
<cmtprojectpath><path value="/extra/xml/dir"/></cmtprojectpath>
</slot></configuration>"""
        }
    ) as tmp:
        os.environ["LHCBNIGHTLIES"] = tmp
        opts, _ = parse_args(addSearchPath, ["--nightly", "lhcb-nightly-slot"])
        assert "/extra/xml/dir" in map(str, opts.dev_dirs)


def test_lookup():
    from LbEnv.ProjectEnv import path

    bk_path = list(path)

    from LbEnv.ProjectEnv.lookup import (
        MissingDataPackageError,
        MissingManifestError,
        MissingProjectError,
        findDataPackage,
        findProject,
        getEnvXmlPath,
    )

    if "BINARY_TAG" in os.environ:
        del os.environ["BINARY_TAG"]
    os.environ["CMTCONFIG"] = "x86_64-slc6-gcc46-opt"

    data = {
        "MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest></manifest>',
        "DERIVEDPROJECT/DERIVEDPROJECT_v1r3/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest>'
        "<used_projects>"
        '<project name="MyProject" version="v1r0" />'
        "</used_projects>"
        "<used_data_pkgs>"
        '<package name="MyPackage" version="*"/>'
        "</used_data_pkgs>"
        "</manifest>",
        "OLDPROJECT/OLDPROJECT_v9r0/InstallArea/x86_64-slc6-gcc46-opt": None,
        "DBASE/MyPackage/v2r0": None,
        "DBASE/MyPackage/v2r1": None,
        "DBASE/MyPackage/v10r0": None,
        "User/MyProjectDev_v1r0/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest>'
        '<used_projects><project name="MyProject" version="v1r0" /></used_projects>'
        "</manifest>",
        "User/MyAnalysis/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest>'
        '<used_projects><project name="MyProject" version="v1r0" /></used_projects>'
        "</manifest>",
    }
    with TempDir(data) as tmp:
        path[:] = [tmp] + bk_path

        proj = findProject("MyProject", "v1r0", os.environ["CMTCONFIG"])
        assert proj == os.path.join(
            tmp, "MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc46-opt"
        )

        try:
            findProject("AnotherProject", "v1r0", os.environ["CMTCONFIG"])
            assert False, "exception expected"
        except MissingProjectError as x:
            assert str(x)

        # test with user area
        path.insert(0, os.path.join(tmp, "User"))
        proj = findProject("MyProjectDev", "v1r0", os.environ["CMTCONFIG"])
        assert proj == os.path.join(
            tmp, "User/MyProjectDev_v1r0/InstallArea/x86_64-slc6-gcc46-opt"
        )

        proj = findProject(
            "MyAnalysis", "latest", os.environ["CMTCONFIG"], allow_empty_version=True
        )
        assert proj == os.path.join(
            tmp, "User/MyAnalysis/InstallArea/x86_64-slc6-gcc46-opt"
        )

        try:
            findProject("MyAnalysis", "v1r0", os.environ["CMTCONFIG"])
            assert False, "exception expected"
        except MissingProjectError as x:
            assert str(x)

        pkg = findDataPackage("MyPackage", "v2r0")
        assert pkg == os.path.join(tmp, "DBASE/MyPackage/v2r0")

        pkg = findDataPackage("MyPackage", "v2r*")
        assert pkg == os.path.join(tmp, "DBASE/MyPackage/v2r1")

        pkg = findDataPackage("MyPackage", "v*")
        assert pkg == os.path.join(tmp, "DBASE/MyPackage/v10r0")

        pkg = findDataPackage("MyPackage", "*")
        assert pkg == os.path.join(tmp, "DBASE/MyPackage/v10r0")

        try:
            findDataPackage("NoPackage", "*")
            assert False, "exception expected"
        except MissingDataPackageError as x:
            assert str(x)

        xml_path = getEnvXmlPath("DerivedProject", "v1r3", os.environ["CMTCONFIG"])
        assert (
            os.path.join(
                tmp, "MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc46-opt"
            )
            in xml_path
        )
        assert (
            os.path.join(
                tmp,
                "DERIVEDPROJECT/DERIVEDPROJECT_v1r3/InstallArea/x86_64-slc6-gcc46-opt",
            )
            in xml_path
        )

        try:
            xml_path = getEnvXmlPath("OldProject", "v9r0", os.environ["CMTCONFIG"])
            assert False, "exception expected"
        except MissingManifestError as x:
            assert str(x)

    path[:] = bk_path


def test_listVersions():
    from LbEnv.ProjectEnv import path

    bk_path = list(path)
    from os.path import join

    from LbEnv.ProjectEnv.lookup import listVersions

    if "BINARY_TAG" in os.environ:
        del os.environ["BINARY_TAG"]
    os.environ["CMTCONFIG"] = "x86_64-slc6-gcc48-opt"

    data = {
        "MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc48-opt/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest></manifest>',
        "MYPROJECT/MYPROJECT_v1r2/InstallArea/x86_64-slc6-gcc48-opt/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest></manifest>',
    }
    with TempDir(data) as tmp:
        path[:] = ["/no/where", tmp, tmp]

        expected = [
            ("v1r2", join(tmp, "MYPROJECT", "MYPROJECT_v1r2")),
            ("v1r0", join(tmp, "MYPROJECT", "MYPROJECT_v1r0")),
        ]
        observed = list(listVersions("MyProject", os.environ["CMTCONFIG"]))
        print("expected", expected)
        print("observed", observed)
        assert expected == observed

    path[:] = bk_path


def _test_profiling():
    from StringIO import StringIO

    from LbEnv.ProjectEnv import profiling

    try:
        sys.stderr = StringIO()
        profiling.run("sys.exit()")
        assert False, "exception expected"
    except SystemExit:
        print(sys.stderr.getvalue())
    finally:
        sys.stderr = sys.__stderr__

    def dummy():
        raise ImportError()

    import cProfile

    bk1 = cProfile.Profile
    cProfile.Profile = dummy

    try:
        sys.stderr = StringIO()
        profiling.run("sys.exit()")
        assert False, "exception expected"
    except SystemExit:
        print(sys.stderr.getvalue())
    finally:
        sys.stderr = sys.__stderr__

    import profile

    bk2 = profile.Profile
    profile.Profile = dummy
    try:
        sys.stderr = StringIO()
        profiling.run("sys.exit()")
        assert False, "exception expected"
    except SystemExit:
        print(sys.stderr.getvalue())
    finally:
        sys.stderr = sys.__stderr__

    cProfile.Profile = bk1
    profile.Profile = bk2


def test_LBCORE_522():
    """LBCORE-522: lb-run does not expand some variables when falling back on SetupProject
    https://its.cern.ch/jira/browse/LBCORE-522"""
    from LbEnv.ProjectEnv.compatibility import expandAllVars

    original = {"MYDATA": "${DATA}/subdir", "DATA": "/main/path"}
    expected = {"MYDATA": "/main/path/subdir", "DATA": "/main/path"}
    assert expandAllVars(original) == expected

    try:
        expandAllVars({"A": "${B}/x", "B": "${A}/y"})
        assert False, "exception expected"
    except ValueError:
        pass


def test_SearchPath():
    from LbEnv.ProjectEnv import EnvSearchPathEntry, SearchPath, SearchPathEntry

    if "DUMMY_ENV_VAR" in os.environ:
        del os.environ["DUMMY_ENV_VAR"]
    try:
        EnvSearchPathEntry("DUMMY_ENV_VAR")
        assert False, "exception expected"
    except ValueError:
        pass

    assert os.environ["PATH"] == str(EnvSearchPathEntry("PATH"))

    path = SearchPath(
        [SearchPathEntry(p) for p in os.environ["PATH"].split(os.pathsep)]
    )
    assert os.environ["PATH"] == str(path)
    assert os.environ["PATH"].split(os.pathsep) == list(path)

    path2 = SearchPath([EnvSearchPathEntry("PATH")])
    assert os.environ["PATH"].split(os.pathsep) == list(path2)

    path3 = SearchPath([SearchPathEntry("b")])
    path3.insert(0, "a")
    path3.append("c")
    assert os.pathsep.join("abc") == str(path3)

    path4 = path3 + list("123")
    assert os.pathsep.join("abc123") == str(path4)

    path5 = path3 + SearchPath("123")
    assert os.pathsep.join("abc123") == str(path5)

    assert not SearchPath([])
    assert SearchPath(["a"])


def test_incompatibility_report():
    from LbPlatformUtils.inspect import os_id

    if not os_id().endswith("centos7"):
        pytest.skip("only centos7 is incompatible with slc5")
    from subprocess import STDOUT, CalledProcessError, check_output

    try:
        check_output(
            [
                "lb-run",
                "--disallow-containers",
                "--platform",
                "x86_64-slc5-gcc43-opt",
                "LHCb/v31r6p1",
                "true",
            ],
            stderr=STDOUT,
        )
        assert False, "script failure expected"
    except CalledProcessError as err:
        print(err.output)
        import re

        assert re.search(rf"dirac_platform: [^-]+-{os_id()}", err.output.decode())


def test_path_to_project():
    import LbEnv.ProjectEnv.lookup
    from LbEnv.ProjectEnv import path
    from LbEnv.ProjectEnv.script import LbRun

    LbEnv.ProjectEnv.lookup.findProject.cache_clear()

    bk_path = list(path)
    path[:] = []

    platform = "x86_64_v2-centos7-gcc11-opt"

    data = {
        f"MyProject_v1r1/InstallArea/{platform}/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest><project name="MyProject" version="1.1" /></manifest>',
        f"MyProject_v1r1/InstallArea/{platform}/MyProject.xenv": '<?xml version="1.0" encoding="UTF-8"?><env:config xmlns:env="EnvSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="EnvSchema EnvSchema.xsd "></env:config>',
    }

    with TempDir(data) as tmp:
        with patch.dict("os.environ", {"CMAKE_PREFIX_PATH": tmp}):
            with patch(
                "sys.argv",
                [
                    "lb-run",
                    "--force-platform",
                    "--platform",
                    platform,
                    "--path-to-project",
                    tmp + "/MyProject_v1r1",
                ],
            ):
                run = LbRun()
                run._makeEnv()
                assert run.env["MYPROJECT_PROJECT_ROOT"] == (tmp + "/MyProject_v1r1")

    path[:] = bk_path


def test_cmake_version_lookup_to_vXrY():
    import LbEnv.ProjectEnv.lookup
    from LbEnv.ProjectEnv import path
    from LbEnv.ProjectEnv.script import LbRun

    LbEnv.ProjectEnv.lookup.findProject.cache_clear()

    bk_path = list(path)
    path[:] = []

    platform = "x86_64_v2-centos7-gcc11-opt"

    data = {
        f"MyProject_v1r1/InstallArea/{platform}/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest><project name="MyProject" version="1.1" /></manifest>',
        f"MyProject_v1r1/InstallArea/{platform}/MyProject.xenv": '<?xml version="1.0" encoding="UTF-8"?><env:config xmlns:env="EnvSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="EnvSchema EnvSchema.xsd "></env:config>',
    }

    with TempDir(data) as tmp:
        with patch.dict("os.environ", {"CMAKE_PREFIX_PATH": tmp}):
            with patch(
                "sys.argv",
                [
                    "lb-run",
                    "--force-platform",
                    "--platform",
                    platform,
                    "--dev-dir",
                    tmp,
                    "MyProject/1.1",
                ],
            ):
                run = LbRun()
                run._makeEnv()
                assert run.env["MYPROJECT_PROJECT_ROOT"] == (tmp + "/MyProject_v1r1")

    path[:] = bk_path


def test_cmake_version_lookup_to_cmake_version():
    import LbEnv.ProjectEnv.lookup
    from LbEnv.ProjectEnv import path
    from LbEnv.ProjectEnv.script import LbRun

    LbEnv.ProjectEnv.lookup.findProject.cache_clear()

    bk_path = list(path)
    path[:] = []

    platform = "x86_64_v2-centos7-gcc11-opt"

    data = {
        f"MyProject_1.1/InstallArea/{platform}/manifest.xml": '<?xml version="1.0" encoding="UTF-8"?><manifest><project name="MyProject" version="1.1" /></manifest>',
        f"MyProject_1.1/InstallArea/{platform}/MyProject.xenv": '<?xml version="1.0" encoding="UTF-8"?><env:config xmlns:env="EnvSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="EnvSchema EnvSchema.xsd "></env:config>',
    }

    with TempDir(data) as tmp:
        with patch.dict("os.environ", {"CMAKE_PREFIX_PATH": tmp}):
            with patch(
                "sys.argv",
                [
                    "lb-run",
                    "--force-platform",
                    "--platform",
                    platform,
                    "--dev-dir",
                    tmp,
                    "MyProject/1.1",
                ],
            ):
                run = LbRun()
                run._makeEnv()
                assert run.env["MYPROJECT_PROJECT_ROOT"] == (tmp + "/MyProject_1.1")

    path[:] = bk_path
